import gc 
import re 
import sys  
import warnings 
import os 
import time  
from datetime import datetime 
import warnings   
import pandas as pd
import numpy as np

import hashlib
from collections import defaultdict, Counter

from .sequence_functions import list_diff_outer_join, lcs, filter_lcs
from .os_functions import *
from .df_functions import *
from .config_table import ConfigReader 
from .excel_functions import write_format_columns, refresh_excel_calculations, save_csv
from .regex_functions import *
from .decorator_functions import *

def replace_brackets2blank(string):
    for x in '()（）-_':
        string = string.replace(x,' ')
    return string 

@catch_and_print
def check_mapping_complete(df_worksheet, complete_header_df, original2cn_dict,file_tag):
    #检查是否遗漏了不是“无”的映射字段`
    c_required_columns = get_dict_unique_values(original2cn_dict)
    check_required_columns = set(c_required_columns) - set(df_worksheet.columns)
    check_required_columns = [x for x in check_required_columns \
                if 'fillbeforeconcat:' not in x.replace(' ','') or '合并前填充:' not in x.replace(' ','') ]

    if check_required_columns:
        warning_msg = 'Warning: Failed to find the the mapping for column ' \
                      + f"{','.join(check_required_columns)}" \
                      + f'of excel file tagging with "{file_tag}"'

@catch_and_print
def dtype_handle(complete_header_df, dtype_dict, output=False):
    if dtype_dict:
        #转换成读取格式,主要处理日期有时读取成float格式的问题
        for dtype_column, dtype_type in dtype_dict.items():
            #处理输入数据的类型 日期类型--不包含时分秒 normalize().date 时间类型-包含时分秒-to_datetime, 和其他pandas支持的数据类型
            dtype_type_temp = dtype_type.strip().lower()
            if dtype_column not in list(complete_header_df.columns):
                enter_exit(f'Cannot find "{dtype_column}" in concatnated_data!')
            if dtype_type_temp == '' or dtype_type_temp == '默认' or dtype_type_temp == 'default':
                pass 
            elif 'date' in dtype_type_temp or 'time' in dtype_type_temp:
                complete_header_df = normalize_dates(complete_header_df, dtype_column)
                if 'date' in dtype_type_temp:  #如果是只要日期，把它转成没有时分秒的格式
                    complete_header_df[dtype_column] = complete_header_df[dtype_column].dt.normalize()
                    #time 不用处理，前面已经normalize_date转成时间格式了
            else:
                #如果是日期，需要转成固定的2020/09/09的格式 提供给技术的java补录工具读取
                if output and dtype_type_temp == 'str' and np.issubdtype(complete_header_df[dtype_column].dtype, np.datetime64):
                    complete_header_df[dtype_column] = complete_header_df[dtype_column].dt.strftime('%Y/%m/%d')
                else:
                    #防止转成空值的float 类型变成 nan, 统一fillna
                    try:  #防止有空值concat后全是float转str 其他非空的int也会带个小数点
                        complete_header_df[dtype_column] = complete_header_df[dtype_column].fillna('')\
                        .astype(dtype_type).apply(lambda x: x.rstrip('.0') if type(x) == str and len(x) >= 2 and x[-2:] == '.0' else x )
                    except :
                        enter_exit(f'Failed to convert column "{dtype_column}" to dtype "{dtype_type}"')

    return complete_header_df

@catch_and_print
def get_dict_unique_values( dictionary):
    """提取一个字典里面的所有非重复的values作为key构建新的字典，新字典values统一为1"""
    record_set = set()
    for k, v in dictionary.items():
        record_set.add(v)
    return record_set

@catch_and_print
def combine_multi_plus(df_worksheet,original2cn_dict):
    #加入系统表格处理,联系方式合并Tel,Email,Facebook
    multi_plus_list =  [x.lower() for x in original2cn_dict.keys() if '+' in x ]
    if multi_plus_list:
        for m in multi_plus_list:
            m_split_list = m.split('+')
            find_lack_columns(df_worksheet, m_split_list,'combine_multi_plus')
            df_worksheet[m] = df_worksheet[m_split_list[0]] 
            for m_s in m_split_list[1:]:
                df_worksheet[m] = df_worksheet[m].fillna('').astype(str).replace('/','')\
                                  + ' ' + df_worksheet[m_s].fillna('').astype(str).replace('/','')
            df_worksheet[m] = df_worksheet[m].str.strip()

    return df_worksheet


@catch_and_print
def get_filter_condition_match(join_table_df, filter_condition):
    #匹配的字段条件和模糊匹配(字段标准化)不同
      
    if filter_condition != '' :
        join_table_df = df_query(join_table_df,filter_condition)
        
    return join_table_df 

@catch_and_print
def is_df_column(string):
    return '[' == string[0] and ']' == string[-1] 

@catch_and_print
def get_filter_condition_standardize_tag(filter_condition):
    #仅限于模糊匹配，里面可能会填入 [国家/地区 == 国家/地区] 的形式
    #代表是否填入的两个表的相同或不同字段
    filter_condition_2_columns_tag = False 
    filter_left_column, filter_right_column = '', ''
    if is_df_column(filter_condition) : 
        filter_condition_2_columns_tag = True 

        filter_condition = filter_condition.lstrip('[').rstrip(']')
        filter_condition_list = [ x.strip() for x in filter_condition.split('==') ]
        if len(filter_condition_list) < 2 :
            enter_exit(f'Standardization:failed to compile filter condition:{filter_condition}!') 
        else:
            filter_left_column = filter_condition_list[0]
            filter_right_column = filter_condition_list[1] 

    return filter_condition_2_columns_tag, filter_left_column, filter_right_column

@catch_and_print
def process_sort_order(join_table_df, join_columns, string):
    if string != '' :
        order_columns = string.split('\n')[0].split(':')
        reverse_order = string.split('\n')[1].split(':')

        reverse_order = [ True if x.lower() == 'desc' else False for x in reverse_order ]

        order_columns_new = [ ]
        reverse_ordere_new = [ ]
        for sort, order in zip(order_columns, reverse_order) :
            #join_columns是用来确保 匹配前的on字段也有参与排序，并且不会和人工填入的字段出现重复
            if sort in join_table_df.columns:
                if sort not in join_columns :
                    order_columns_new.append(sort)
                    reverse_ordere_new.append(order)
            else:
                print(f'Match table cannot find column "{sort}"')

        order_columns_new = join_columns + order_columns_new
        reverse_ordere_new = [ True for x in range(len(join_columns)) ]  + reverse_ordere_new

        join_table_df = join_table_df.sort_values(by=order_columns_new, ascending=reverse_ordere_new )

        join_table_df = join_table_df.fillna('').astype(str)
        
    return join_table_df

@catch_and_print
def calc_similarity(not_standard_str, new_standard_str,calc_func ):

    #先把括号换成空格
    not_standard_str = replace_brackets2blank(not_standard_str)
    new_standard_str = replace_brackets2blank(new_standard_str)

    str_similarity = 0 

    match_list = calc_func(not_standard_str, new_standard_str.strip())
    #如果这里就发现一个都没有发现匹配到，直接填相似度为0
    if len(match_list) < 2:
        return str_similarity

    #未处理符号的连续字母命中最多的优先,'PD1945F/FF/CF/DF/EF'， 'PD1945BF/DF/FF_EX' PD1945F  应该优先选第一个，第二个只是长度短而已
    continues_match = 0 
    for i in range(len(not_standard_str)):
        try:
            if not_standard_str[i] == new_standard_str[i]:
                continues_match += 1 
            else:
                break
        except:
            break

    #解决上面有些因为内存而导致匹配到的字符长，但却匹配错的问题
    not_standard_str_no_space = ' '.join(not_standard_str.split(' ')[:-1])
    new_standard_str_no_space = ' '.join(new_standard_str.split(' ')[:-1])
    #用分母的方式，解决 Y15 优先 配到了 Y15S 而不是 Y15(4+64G)的问题
    match_list_no_space = calc_func(not_standard_str_no_space, new_standard_str_no_space )
    
    match_list_len = len(match_list)
    match_list_len_no_space = len(match_list_no_space) 


    try: #也门 -- 对应被拆开的 卡塔尔 也门 科威特 阿曼,第一个卡塔尔拆出来之后 分母为0 
        match_list_len_no_space_divide = len(not_standard_str_no_space) / len(new_standard_str_no_space)
    except ZeroDivisionError:
        match_list_len_no_space_divide = 0 

    str_similarity = match_list_len + continues_match + match_list_len_no_space + match_list_len_no_space_divide

    return str_similarity

@catch_and_print
def replace_by_dict(string, replace_dict={}):
    if replace_dict :    
        #先做替换
        for k,v in replace_dict.items():
            string, sub_num = re.subn(k, v, string, flags=re.I)
            #如果已经替换成功就停止，不往下替换
            if sub_num != 0 :
                break
    return string 

@catch_and_print
def check_similarity_number( not_standard_str, standard_dict,special_syn_list,
                             ignore_punctuation=True):

    #某些情况不符合模糊匹配条件默认加入的结果 
    not_match_default = [0, '', '']

    first_syn = [ x for x in special_syn_list if x != '' ][0]

    for standard_str, target_str in standard_dict.items():
        #创建一个专门用来匹配的新standard_str
        new_standard_str = standard_str.lower().strip()

        #和其他两个匹配方式有区别，忽略掉ignore_punctuation为了保证regex的结果
        if special_syn_list:
            syn_checking = check_syn_str_regex_number(not_standard_str,new_standard_str,special_syn_list)
            #如果检查两边两边的同步字符不同, 直接判断为不符合
            if not syn_checking:
                continue 
        #需要填入ram_rom 获取到的部分 才能对内存数字是否相同做到完美的判断, 不放入first_syn会导致 (y3 16G+128)可能会匹配到(3+16G)
        match_list = number_similarity(not_standard_str,new_standard_str, first_syn)

        #以下可以直接用是否返回match_list来判断
        if match_list :
            return [[99, new_standard_str, target_str]]

    return [ not_match_default ]

@catch_and_print  
def compare_number_part(string1, string2):
    not_standard_number = ''
    new_standard_number = ''

    counter = 0 
    for s in string1:
        if s.isdigit():
            counter += 1 
            not_standard_number += s 
        else:
            if counter >= 1 : 
                break 
    counter = 0 
    for s in string2:
        if s.isdigit():
            counter += 1 
            new_standard_number += s 
        else:
            if counter >= 1 :
                break

    return not_standard_number ==  new_standard_number

@catch_and_print
def check_similarity_simple( not_standard_str, standard_dict,special_syn_list,
                             ignore_punctuation=True):
    """
    1. 完全相等直接返回
    2. 需要两边同时有special_syn_list, 同时ignore_punctuation
    3. 相同的字符至少要有两个及以上
    4. 从遇到第一个数字开始连续的数字部分（第一个数字部分）必须相同, PD1945不能因为没有别的相似而配上PD1948
    """
    #not_standard_str不要填入空字符串防止运行累赘,在前面的函数做判断
    standard_similarity_list = []

    #某些情况不符合模糊匹配条件默认加入的结果 
    not_match_default = [0, '', '']

    for standard_str, target_str in standard_dict.items():
        #创建一个专门用来匹配的新standard_str
        new_standard_str = standard_str.lower().strip()
       #确认至少两个字符相同，并且如果出现数字，从左往右遇到第一个完整连续的数字串，必须完全相同
        check_number_part = compare_number_part(not_standard_str, new_standard_str)
        if not check_number_part:
            standard_similarity_list.append(not_match_default)
            continue  

        #如果完全相等直接,不往下面走, 直接返回
        if not_standard_str == new_standard_str:
            return [[99, standard_str, target_str]]

        if ignore_punctuation == True:
            new_standard_str = replace_punctuations(new_standard_str)

        if special_syn_list:
            syn_checking = check_syn_str_regex(not_standard_str,new_standard_str,special_syn_list)
            #如果检查两边两边的同步字符不同, 直接返回判断为
            if not syn_checking:
                standard_similarity_list.append(not_match_default)
                continue

        str_similarity = calc_similarity(not_standard_str, new_standard_str, lcs )

        #上面一截 除了lcs函数不同，其他都一样
        if str_similarity >=2  :
            standard_similarity_list.append([str_similarity, standard_str, target_str])
        else:
            standard_similarity_list.append(not_match_default)

    standard_similarity_list = sorted(standard_similarity_list, key=lambda x:x[0],reverse =True)

    return standard_similarity_list

@catch_and_print
def check_similarity_strict( not_standard_str, standard_dict,special_syn_list,
                             replace_dict={}, ignore_punctuation=True):
    #not_standard_str不要填入空字符串防止运行累赘
     #机型匹配模式，用的是filter_lcs,因为要确保前面2个或3个字母都相同
    standard_similarity_list = []

    #某些情况不符合模糊匹配条件默认加入的结果 
    not_match_default = [0, '', '']

    for standard_str, target_str in standard_dict.items():
        #创建一个专门用来匹配的新standard_str
        new_standard_str = standard_str.lower().strip()

        #如果完全相等直接,不往下面走, 直接返回
        if not_standard_str == new_standard_str:
            return [[99, standard_str, target_str]]

        if ignore_punctuation == True:
            new_standard_str = replace_punctuations(new_standard_str)

        if special_syn_list:
            syn_checking = check_syn_str_regex(not_standard_str,new_standard_str,special_syn_list)
            #如果检查两边两边的同步字符不同, 直接返回判断为
            if not syn_checking:
                standard_similarity_list.append(not_match_default)
                continue

        str_similarity = calc_similarity(not_standard_str, new_standard_str,filter_lcs )

        if str_similarity >= 2 : 
            # 有3个和3个以上的字符，前3个全是字母确保3个字母都相同，前3个字符有数字和英文，确保英文和数字都相同
            # 前两/三位字母相同,方便后面elif写条件
            first2_letter_equal = not_standard_str[:2] == new_standard_str[:3].split(' ')[0][:2]
            first3_letter_equal = not_standard_str[:3] == new_standard_str[:3].split(' ')[0][:3]

            #前3位字符串包含2个数字的情况 必须满足前3个字母相同
            first3_letter_2num_not_standard = re.search('[0-9]{2,}',not_standard_str[:3]) != None
            first3_letter_2num_standard =  re.search('[0-9]{2,}',new_standard_str[:3]) != None

            first3_letter_less2num_not_standard = not first3_letter_2num_not_standard
            first3_letter_less2num_standard = not first3_letter_2num_standard

            #1. 如果输入机型只有2个字符，空格分割开的字符必须完全相同: 因为V1不能配成V11
            if len(not_standard_str) == 2 :
                new_standard_str = new_standard_str.split(' ')[0]

                if not_standard_str == new_standard_str:
                    standard_similarity_list.append([str_similarity, standard_str, target_str])
                else:
                    standard_similarity_list.append(not_match_default)

            elif   (len(not_standard_str) >= 3 and first2_letter_equal) and \
             (     #输入字符 前3位有2个数字 + 标准字符 前3位也有2个数字 -- > 前3个必须要相等
                   (first3_letter_2num_not_standard and first3_letter_2num_standard and first3_letter_equal) \
                   #输入字符 前3位没有2个数字 + 标准字符 前3位也没有2个数字 -- > 前2个字符必须要相等
                or (first3_letter_less2num_not_standard and first3_letter_less2num_standard and first2_letter_equal)
             ):
                standard_similarity_list.append([str_similarity, standard_str,target_str])
            else :
                standard_similarity_list.append(not_match_default)
        else:
            standard_similarity_list.append(not_match_default)

    standard_similarity_list = sorted(standard_similarity_list, key=lambda x:x[0],reverse =True)

    return standard_similarity_list

@catch_and_print
def standardize_by_similarity( not_standard_str, standard_dict, special_syn_list,
                               replace_dict={},  mode ='filter_lcs',ignore_punctuation=True):
    #mode, simple_lcs --用来匹配简单的国家 or filter_lcs -- 用来匹配机型
    #替换干扰符号和干扰词
    #处理中文国家缩写和完整国家名称无法匹配到的情况

    #前面已经做过去重
    not_standard_str = replace_by_dict(not_standard_str, replace_dict)

    not_standard_str = not_standard_str.lower().strip()

    standard_dict = dict(sorted(standard_dict.items(), key=lambda x:len(x[0]), reverse= False))

    if mode == 'simple_lcs':
        standard_similarity_list = check_similarity_simple(not_standard_str, standard_dict,special_syn_list,
                                                            ignore_punctuation)
    elif mode == 'number_similarity':
        standard_similarity_list = check_similarity_number(not_standard_str, standard_dict,special_syn_list,
                                                            ignore_punctuation)
    else:
        standard_similarity_list = check_similarity_strict(not_standard_str, standard_dict,special_syn_list,
                                                            ignore_punctuation)

    return standard_similarity_list

@catch_and_print
def standardize_column_func( not_standard_str, standard_dict, special_syn_list,
                             replace_dict={}, mode ='filter_lcs', ignore_punctuation=True):

    standard_similarity_list = standardize_by_similarity(not_standard_str, standard_dict, special_syn_list,
                                                         replace_dict, mode,ignore_punctuation)

    #确保输入的不为空
    if not_standard_str != '' and not_standard_str ==  not_standard_str :
        if standard_similarity_list and standard_similarity_list[0][0] > 0 :
            return standard_similarity_list[0][2]
        else:
            return ''
    
    return ''

@catch_and_print
def dropping_not_mapping(df_worksheet, original2cn_dict, target_cn_columns):
    mapping_column_list = [ x[1] for x in  original2cn_dict.items()]
    #如果现有的字段和需要的target字段有重复，但其实并不想映射进目的字段（因为可能数据是不同含义），先要删掉原始表的重复字段
    for d in df_worksheet.columns: 
        if d in target_cn_columns and d not in mapping_column_list:
            df_worksheet = df_worksheet.drop(d, axis=1) 
    return df_worksheet

@catch_and_print
def read_csv_data(table_path):
    df_worksheet = pd.DataFrame([])
    for encoding in ['utf-8','gbk','gb2312'] :
        try:
            df_worksheet = pd.read_csv(table_path, encoding=encoding )
            break
        except:
            print('')
            print(f'***** Failed to read file:{table_path} with encoding: {encoding} ***** ')
            print('')
            continue
    if df_worksheet.empty:
        print(f'***** Failed to read file:{table_path} *****')
    return df_worksheet

@catch_and_print
def read_xls_special(table_path):
    df_worksheet = pd.DataFrame([])
    try:
        df_worksheet = pd.read_html(table_path,header=0)[0]
    except:
        print('')
        print(f'***** Failed to read file:{table_path} ***** ')
        print('')

    return df_worksheet 

@catch_and_print
def get_min_max_date(result_df ):
    min_max_date_range = ''
    date_time_df = result_df.select_dtypes(include=['datetime'])

    if not date_time_df.empty:
        datetime_series = date_time_df.iloc[:,0]
        #报错的情况：所有值都是nan  ['NaT' 'NaT' 'NaT' ... 'NaT' 'NaT' 'NaT']
        try:
            min_datetime = datetime.datetime.strftime(datetime_series.min(), '%Y.%m.%d')
            max_datetime = datetime.datetime.strftime(datetime_series.max(), '%Y.%m.%d')
            min_max_date_range = f'{min_datetime}-{max_datetime}'
        except:
            pass
    return min_max_date_range

@catch_and_print
def drop_duplicated_columns_before_rename(df_worksheet, original2cn_dict):
    original_columns = list(df_worksheet.columns)

    duplicated_columns = [ ]
    for k, v in original2cn_dict.items():
        if v in original_columns:
            duplicated_columns.append(v)

    df_worksheet = df_worksheet.drop(duplicated_columns,axis=1)
    return df_worksheet

@catch_and_print
def refresh_configs(config_table_path_list):
    #读取配置前，是否刷新一遍文档内的公式并保存，注意这里不能打开配置文件，否则会无法保存成功
    for c in config_table_path_list:
        print(f'Refreshing excel:{c}')
        refresh_excel_calculations(c)

@catch_and_print
def get_standard_mode(standardize_mode):
    if 'number' in standardize_mode.lower():
        mode = 'number_similarity'
    elif 'simple' in standardize_mode.lower():
        mode = 'simple_lcs'
    else:
        mode = 'filter_lcs'
    return mode 

@catch_and_print
def get_partial_not_match(complete_header_df_notna, seq,source_column, 
                            standard_table_name, standard_column, 
                           filter_condition, target_column_edit, filter_left_column):

    column_order = ['Standization No.','Source_column', 'Standard_column', 'Content needs to standardized', 'Filter condition','Filter content']

    stand_column = '--'.join([standard_table_name,standard_column])

    #获取模糊匹配(标准化)失败的部分
    find_lack_columns(complete_header_df_notna, [target_column_edit, source_column])

    temp_columns = [source_column ]
    temp_headers = ['Content needs to standardized']
    if filter_left_column != '':
        temp_columns.append(filter_left_column)
        temp_headers.append('Filter content')

    partial_match_not_match_df = complete_header_df_notna\
            .loc[(complete_header_df_notna[target_column_edit].isna()==True)\
            |(complete_header_df_notna[target_column_edit]==''), temp_columns ] 

    if not partial_match_not_match_df.empty:
        #获取到的原始列内容
        partial_match_not_match_df.columns = temp_headers

        partial_match_not_match_df['Source_column']  = source_column
        partial_match_not_match_df['Standard_column'] = stand_column
        partial_match_not_match_df['Filter condition'] = filter_condition

        partial_match_not_match_df['Standization No.'] = seq

    partial_match_not_match_df = func_loc(partial_match_not_match_df, column_order) 

    partial_match_not_match_df = partial_match_not_match_df.drop_duplicates()

    return partial_match_not_match_df


@catch_and_print
def process_duplicates( df, drop_duplicates_condition):
    #字段去重
    if type(drop_duplicates_condition) == str and drop_duplicates_condition.strip() != '':

        drop_columns = drop_duplicates_condition.split(':')
        lack_column_list = find_lack_columns(df, drop_columns)
        df = df.drop_duplicates(subset = drop_columns)

    return df 

@catch_and_print
def replace_value_func(complete_header_df_new, replace_value_str, input_column, output_column):
    if replace_value_str != '':
        #统一变成字符型类
        replace_values = replace_value_str.split('\n')

        for replace_value in replace_values:

            #获取到需要替换的原始值和替换值
            replace_list = replace_value.rsplit(':',1)
            if len(replace_list) > 1 :
                #允许替换前的字段内容包含有冒号,从右开始拆分
                replace_original = replace_list[0]
                replace_result = replace_list[1]
            else:
                replace_original = ''
                replace_result = replace_list[1]

            #如果原始内容非空 并且 需要替换成单纯的其他字符串
            if replace_original != '' and 'regex(' != replace_original[:6].lower() :  #支持regex替换
                complete_header_df_new[output_column] = complete_header_df_new[output_column]\
                        .str.replace(replace_original,replace_result,regex=False)

            elif replace_original != '' and 'regex(' == replace_original[:6].lower():
                replace_original = replace_original[6:-1]

                complete_header_df_new[output_column] = complete_header_df_new[output_column]\
                        .str.replace(replace_original,replace_result,case=False, regex=True)
                complete_header_df_new[output_column] = complete_header_df_new[output_column]\
                        .apply(lambda x: re.sub(replace_original,'', x, flags=re.I)).copy()
            else: #如果原始内容为空，即特别需要填充空值,且不能用series.str
                complete_header_df_new[output_column] = complete_header_df_new[output_column]\
                        .fillna(value=replace_result).replace('',replace_result,regex=False)

    return complete_header_df_new


@catch_and_print
def fillna_value_func(complete_header_df, complete_header_df_new, fillna_value_str, output_column):

    if fillna_value_str != '' :
        #如果不是用别的字段做填充
        if not is_df_column(fillna_value_str):
            complete_header_df_new[output_column] = complete_header_df_new[output_column]\
                                .fillna(value=fillna_value_str).replace('',fillna_value_str)
        #如果是属于表格的字段作为填充
        else :
            fillna_value_str_strip = fillna_value_str.lstrip('[').rstrip(']')
            #注意" 在前面整个字段可能已经被转成string格式，不能单纯通过fillna的方式填充空值,
            #而replace的方式又无法把其他字段值作为入参
            if fillna_value_str_strip in complete_header_df_new.columns :
                complete_header_df_new[output_column] = complete_header_df[output_column]\
                .replace('',float('nan')).fillna(value=complete_header_df_new[fillna_value_str_strip])
            else:
                try:  #想填充的可能是来自旧complete_header_df的字段
                    complete_header_df_new[fillna_value_str_strip] = complete_header_df[fillna_value_str_strip]

                    complete_header_df_new[output_column] = complete_header_df_new[output_column]\
                    .replace('',float('nan')).fillna(complete_header_df_new[fillna_value_str_strip])
                                        
                except KeyError:
                    enter_exit(f"Fillna error: Column '{fillna_value_str_strip}' doesn't exist!")

                complete_header_df_new = complete_header_df_new.drop(fillna_value_str_strip, axis=1)

    return complete_header_df_new


@catch_and_print
def sort_value_func(complete_header_df_new, sort_column_list,sort_column_order_list):
    #通过数字来排列数字
    order_list = [ ]
    for column,order in zip(sort_column_list,sort_column_order_list):
        order_list.append((column, order))

    order_list = sorted(order_list,key=lambda x: abs(x[1]))

    sort_column_list = [ x[0] for x in order_list ]
    ascending_list = [ True if x[1] >= 0 else False for x in order_list ]

    try:
        complete_header_df_new = complete_header_df_new.sort_values(by=sort_column_list,ascending=ascending_list)
    except :
        print('Fill&Order:Error when sorting the values')
        
    return complete_header_df_new


@catch_and_print
def check_mapping_duplicates(df, target_cn_columns, table_stem=''):
    #检查映射后是否出现重复字段,  返回需要保留的映射字段即可
    used_cols = [x for x in df.columns if x in target_cn_columns]

    # 检查有没有出现重复字段
    count_duplicate = Counter(used_cols).most_common(1)

    if count_duplicate and count_duplicate[0][1] > 1:
        enter_exit('Table "{}" has {} original fields can be mapped to target field "{}" that leads to duplication!'
                   .format(table_stem, count_duplicate[0][1], count_duplicate[0][0]))

    return df 

@catch_and_print
def get_replace_dict(string):
    try:
        replace_dict = [x.strip()for x in string.split('\n') if x.strip() != '']
        replace_dict = dict([[y.strip() for y in split_colon(x) ] for x in replace_dict])
    except:
        replace_dict = {}

    return replace_dict


@catch_and_print
def get_save_name(output_dir,config_file_dir, min_max_date_range):

    #获取config后面的一截用来做output文件的部分名称
    save_name = re.split('[_\-]',config_file_dir,maxsplit=1)[-1]
    
    if min_max_date_range != '':
        save_name = save_name + '-' +  min_max_date_range

    save_name = os.path.join(output_dir, save_name)

    return save_name

@catch_and_print
def get_save_sheet_name(config_table_name):

    sheet_name = re.sub(pattern='(.*)(\.xlsx|\.csv|.xls)', repl='\g<1>', string=config_table_name)
    sheet_name = re.split('[_\-]',sheet_name)[-1]

    #重复符号换成一个
    sheet_name = replace_multi_symbol(sheet_name,'-')

    sheet_name = strip_puntuations(sheet_name)

    if sheet_name == '':
        sheet_name = f'result{counter}'

    return  sheet_name


@catch_and_print
def save_results(save_name, result_df_list, sheet_name_list,
                 checking_result_list, checking_result_sheet_names, output_file_type):
    
    #确保output目录存在，否则创建output目录
    check_create_new_folder(f'{save_name}.xlsx')

        #CSV文档只针对单个输出结果，多个输出结果的话统一用excel保存
    if type(output_file_type) == str and output_file_type.lower() == 'csv':
        if len(result_df_list) == 1 :
            save_csv(result_df_list[0],f'{save_name}.csv' )
        else: #如果有多个结果，需要分开写入CSV
            for result_df,sheet_name in zip(result_df_list, sheet_name_list):
                save_csv(result_df,f'{save_name}({sheet_name}).csv' )

    elif len(result_df_list) == 1 and result_df_list[0].shape[0] > 100 * 10000 :
        print('The total number of rows exceeds 1 million, result will be saved in CSV format.')
        save_csv(result_df[0], f'{save_name}.csv')

    else:
        write_format_columns(f'{save_name}.xlsx',result_df_list, sheet_name_list, min_column_width=14)

    #写入检查数据
    if checking_result_list:
        path_stem = Path(f'{save_name}.xlsx').stem
        write_format_columns(os.path.join('.\\result_checking', f'Checking results-{path_stem}.xlsx'),
                             checking_result_list, checking_result_sheet_names, min_column_width=16)
            

def split_colon(string):
    lst = [ x for x in string.replace('：', ':').split(':')  if x != '' ]
    return lst

    
def process_join_table(join_table_df, join_columns, target_columns,filter_condition,sort_order, join_table_name = ''):
    #防止里面有重复字段
    join_table_df = remove_duplicate_columns(join_table_df)

    # 获取过滤条件后的关联表
    if filter_condition  != '':
        join_table_df = get_filter_condition_match(join_table_df, filter_condition)

    #定位join_table_df中所有需要的字段.不需要的字段删掉
    temp_used_cols = list(set(join_columns + target_columns))
    join_table_df = join_table_df.loc[:, temp_used_cols]

    # 将所有关联列转成string格式, 这里截取了固定的字段
    #需要对关联的所有字段清楚掉空值，否则如果匹配表填的有问题 会出现匹配错误
    join_table_df = join_table_df.apply(lambda x: 
        x.fillna('').astype(str).str.strip().replace('',float('NaN')) if x.name in join_columns else x )

    join_table_df = join_table_df.dropna(subset=join_columns, how='any')

    #剩下如果是空 直接返回空匹配表
    if not join_table_df.empty:

        if sort_order != '':
            join_table_df = process_sort_order(join_table_df, join_columns, sort_order)

        #如果不是用原始表做关联表的话，关联之前必须做去重，避免笛卡尔积现象
        if  join_table_name != '':
            join_table_df = join_table_df.drop_duplicates(subset=join_columns)

    return join_table_df

def process_match_complete_table(complete_header_df,source_columns, target_columns , join_columns, join_table_name):
    #转成str
    for s in source_columns:
        complete_header_df[s] = complete_header_df[s].fillna('').astype(str)
    #如果原始表格包含了需要获取到的字段，删掉
    for t in target_columns:
        if t not in source_columns and t in complete_header_df.columns:
            complete_header_df = complete_header_df.drop(target_columns,axis=1)

    return complete_header_df

def check_only_one_match_column(join_table_df, join_columns, target_columns ):
    only_one_match_column = False
    if len(join_columns) == 1 and len(target_columns) == 1 and join_columns[0] == target_columns[0]:
        only_one_match_column = True
        join_table_df['additional_temp'] = join_table_df[target_columns[0]]
        join_columns = join_columns + ['additional_temp']
    else:
        join_columns = join_columns + [target_columns]

    return join_table_df, only_one_match_column


import json


class DfDict(dict):
    #辅助group_word_sum创建的类
    def sum(self ):
        return sum(self.values())

    def __add__(self,other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                result[k] = v + other[k]
            else:
                result[k] = v

        for k, v in other.items():
            if k not in self.keys():
                result[k] = v

        return result

    def __sub__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                result[k] = v - other[k]
            else:
                result[k] = v

        for k, v in other.items():
            if k not in self.keys():
                result[k] = -v

        return result

    def __mul__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                result[k] = v * other[k]
            else:
                result[k] = 0

        for k, v in other.items():
            if k not in self.keys():
                result[k] = 0

        return result

    def __truediv__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                if other[k] != 0:  #如果为0 不输出结果
                    result[k] = v / other[k]
            else:
                result[k] = float(0)

        for k, v in other.items():
            if k not in self.keys():
                result[k] = float(0)

        return result

    def __floordiv__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                if other[k] != 0:  #如果为0 不输出结果
                    result[k] = v // other[k]
            else:
                result[k] = float(0)

        for k, v in other.items():
            if k not in self.keys():
                result[k] = float(0)

        return result

    def __mod__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                if other[k] != 0:  #如果为0 不输出结果
                    result[k] = v % other[k]
            else:
                result[k] = 0

        for k, v in other.items():
            if k not in self.keys():
                result[k] = float(0)

        return result

    def __pow__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                if other[k] != 0:  #如果为0 不输出结果
                    result[k] = v ** other[k]
            else:
                result[k] = 0

        for k, v in other.items():
            if k not in self.keys():
                result[k] = 0

        return result

    def __and__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, 0) != 0 and v != 0 :
                result[k] = True
            else:
                result[k] = False

        for k, v in other.items():
            if k not in self.keys() :
                result[k] = False
            else:
            	result[k] == True

        return result

    def __or__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, 0) == 0 and k == 0 :
                result[k] = False
            else:
                result[k] = True

        for k, v in other.items():
            if k not in self.keys() and k == 0 :
                result[k] = False 
            else:
            	result[k] = True

        return result

    def __xor__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if (other.get(k, 0) == 0 and k == 0) or (other.get(k,0) != 0 and k != 0 ) :
                result[k] = True
            else:
                result[k] = False

        for k, v in other.items():
            if (k not in self.keys() and k == 0) :
                result[k] = False 
            else:
            	result[k] = True

        return result

    def __pow__(self, other):
        result = DfDict({ })
        for k, v  in self.items():
            if other.get(k, None) != None:
                if other[k] != 0:  #如果为0 不输出结果
                    result[k] = v ** other[k]
            else:
                result[k] = 0

        for k, v in other.items():
            if k not in self.keys():
                result[k] = 0

        return result

    def __repr__(self):
        return json.dumps(self)

    def __str__(self):
        return json.dumps(self)



from openpyxl import Workbook as openpyxl_workbook
from openpyxl.styles import Color,PatternFill,Side, Font, Border,Alignment
from openpyxl.utils import get_column_letter
from collections import defaultdict, Counter
from xlsxwriter.exceptions import FileCreateError
from xlsxwriter import Workbook
from .os_functions import check_create_new_folder
from .sequence_functions import duplicate_elem_add_seq
from decimal import Decimal
import datetime 
import re 
import os 
import math 
import warnings 
import numpy as np
import xlwings as xw 

from pandas.core.indexes.multi import MultiIndex

"""pandas save_excel """
def pd_save_excel(df,save_path,startrow=0,startcol=0):

	close_tag = 0
	while close_tag <= 0:
		try:
			df.to_excel(save_path,index=False,startrow=startrow,startcol=startcol)
			close_tag += 1 
		except (PermissionError,FileCreateError) as e :
			input('\nFailed to write file!\n  Please Close "{}" Then Press Enter to Continue'.format(save_path))
	print('{0} Saved'.format(save_path))

"""这里主要是xlsxwriter的方法"""
def autofit_column_width(xlsxwriter_ws,header_columns,content_columns=[],wrap_text=False, **kwargs):
	"""根据表头的字符长度自动调整EXCEL表头的列宽
	   content_columns是专门给特长的列加宽
	"""
	min_column_width = kwargs.get('min_column_width',10)

	cn_pat = "[\u4e00-\u9fa5]+"
	en_pat = "[^\u4e00-\u9fa5]+"

	#先统一把表头转成文字格式
	header_columns = [str(x) for x in header_columns]

	if wrap_text == False:
		length_list = [ len(''.join(re.findall(cn_pat,c)))*2 + len(''.join(re.findall(en_pat,c)))*1  + 4 \
							for c in list(header_columns)] 
		for i, width in enumerate(length_list):
			xlsxwriter_ws.set_column(i,i,width)
		#防止最后一列没修改成功
		xlsxwriter_ws.set_column(i+1,i+1,width)

	else: #如果采用wrap_text方式，计算两层的中文wrap结果 
		length_list = [ math.ceil(len(''.join(re.findall(cn_pat,c)))/2) * 2 + \
						math.ceil(len(''.join(re.findall(en_pat,c)))/2) * 1 + 2 \
							for c in list(header_columns)] 

		#确保每列的宽度至少是4 
		length_list = [x if x >= min_column_width else min_column_width for x in length_list ]

		for i, width in enumerate(length_list):
			xlsxwriter_ws.set_column(i,i,width)
		#防止最后一列没修改成功
		xlsxwriter_ws.set_column(i+1,i+1,width)

	#给特别的内容列加宽
	if content_columns:
		for c in content_columns:
			c_index = header_columns.index(c)
			xlsxwriter_ws.set_column(c_index,c_index,80)

def adjust_pct_decimal_format(original_number_format,data):
	#如果某一行/列的数据数字50%及以上的部分到第二个百分点无法区分上升或下降,比如0.001%, 0.002%
	#保留2个小数点,四舍五入都是约等于0.00%,则往下一级去扩展保留更多的小数点位数
	original_decimal = original_number_format[:-1].split('.')

	if original_decimal :
		original_decimal_places = len(original_decimal[1]) 
	else:
		original_decimal_places = 2

	#找到该列数据小数点最长的数字的小数点位数
	biggest_decimal_places = max([abs(Decimal(str(x)).as_tuple().exponent) for x in data if type(x)==float ])

	#当约到某一位的数字超过50%时，加一个小数位
	data_decimal_check = [ round(x,original_decimal_places+2) for x in data if type(x)==float ]

	#众数的数量如果超过50%
	mode_counter = Counter(data_decimal_check).most_common()[0]
	mode = mode_counter[0]
	mode_counter = mode_counter[1]
	
	mode_decimal_places = abs(Decimal(mode).as_tuple().exponent)

	#不能把original_decimal_format直接赋值给next_decimal_format,不可修改
	next_decimal_format = original_number_format

	if  biggest_decimal_places > original_decimal_places and mode_counter/len(data_decimal_check) >= 0.5:
		next_decimal_format = '0' * (original_decimal_places + 1)
		next_decimal_format = '0.{}%'.format(next_decimal_format)

		next_decimal_format = adjust_pct_decimal_format(next_decimal_format,data)

	return next_decimal_format

def write_row_format(xlsxwriter_ws,data,start_pos,cell_format={},direction=0):
	"""fit in row or col parameter to write_row method according to direction param
	:param row : start row zero indexed 
	:param col : start col zero indexed
	:param pct_format : how to display num_format
	:param direction : 1/horizontal ，0/vertical
	"""
	data = [ x if x == x else '' for x in data]
	if direction == 1:
		xlsxwriter_ws.write_column(row=0,col=start_pos,data=data,cell_format=cell_format)
	else :
		xlsxwriter_ws.write_row(row=start_pos,col=0,data=data,cell_format=cell_format)

def write_multi_tables(xlsxwriter_wb,sheet_name,df_list,direction=0,**kwargs):
	"""将多个小型数据表写入一个sheet里面
	   0/horizontal 从上往下顺序写入，1/vertical 从左往右顺序写入
	   有额外的kwargs参数，方便输入遇到哪个字符串就用百分比格式
	"""
	percent_str = kwargs.get('percent_str','占比')
	auto_adjust_pct =  kwargs.get('auto_adjust_pct',False)
	normal_format =  kwargs.get('normal_format',{'font_name':'calibri','font_size':11})
	date_format = kwargs.get('date_format','%Y-%m-%d')

	#通用表头格式
	background_color = rgb_convert_hex([220,230,241])
	border_color = rgb_convert_hex([149,179,215])

	header_format = xlsxwriter_wb.add_format({'font_name':'微软雅黑','bold':False,'font_size':11,
											  'bg_color':background_color })
	header_format.set_bottom()
	header_format.set_bottom_color(border_color)

	normal_format = xlsxwriter_wb.add_format(normal_format)

	xlsxwriter_ws = xlsxwriter_wb.add_worksheet(sheet_name)

	#记录表头
	horizontal_columns = [ ]
	vertical_columns = [ ]
	#记录每次位置
	position_counter = 0

	for df in df_list:
		#如果其中某个或所有index不是常规的数字, reset_index
		if not all([type(x) == int for x in list(df.index) ]) :
			df =  df.reset_index()

		#记录横着写的表格所有表头
		horizontal_columns += list(df.columns) + [' ' * 8]
		vertical_columns = list(df.columns)

		#先判断这个表格是否要按照百分比格式写入
		if percent_str in df.columns.values[0]: 
			pct_tag = True
		else: 
			pct_tag = False 

		#表头reset
		df = df.T.reset_index().T.reset_index(drop=True)
		#如果是以列的方式写入
		if direction == 1:
			data_set = [ (i,list(df[x])) for i,x in enumerate(df.columns)]
		else:
			data_set = df.iterrows()

		is_header = True
		for i,data in data_set:
			if is_header:
				#写入表头
				header = list(df.columns.values)
				write_row_format(xlsxwriter_ws,data=data,start_pos=position_counter,
								cell_format=header_format,direction=direction)
				position_counter += 1
				is_header = False 

			else:
				# 如果是日期，转成对应的日期格式
				data = [ datetime.datetime.strftime(x,date_format) if isinstance(x, datetime.datetime) else x for x in data ]
				data = [ float(x) if type(x) != str else x for x in data ]

				if pct_tag:
					#百分比需要判断放入多少个小数点的百分比,不在这里加入
					number_format = kwargs.get('num_format','0.00%')
					if auto_adjust_pct :  #
						number_format = adjust_pct_decimal_format(number_format,data)

					pct_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':number_format})

					write_row_format(xlsxwriter_ws,data=data,start_pos=position_counter,
									 cell_format=pct_format,direction=direction)
				else:
					write_row_format(xlsxwriter_ws,data=data,start_pos=position_counter,
									cell_format=normal_format, direction=direction)

				position_counter += 1 

		#不同表格换行
		position_counter += 1

		if direction == 1:
			autofit_column_width(xlsxwriter_ws,horizontal_columns)
		else:
			autofit_column_width(xlsxwriter_ws,vertical_columns)


def df_sheet_check(df_list,sheet_name_list):
	#防止填入的不是列表格式,防止填入的sheet_name 出现重复
	if type(df_list) != list:
		df_list = [df_list]
	if type(sheet_name_list) != list:
		sheet_name_list = [sheet_name_list]
	
	if len(df_list) != len(sheet_name_list):
		print('写入的数据表数量和填入的Sheet名称数量不相等,将使用默认Sheet名称')

	sheet_name_list = duplicate_elem_add_seq(sheet_name_list)

	if sheet_name_list == None or sheet_name_list == [ ] :
		sheet_name_list= []
		for i in range(len(df_list)):
			sheet_name_list.append('Sheet {}'.format(i+1))

	return df_list, sheet_name_list
	
def write_pct_columns(save_path,df_list,sheet_name_list=[],pct_columns=[],content_columns=[],**kwargs):
	"""给定一个表格，用百分比写入包含指定文字的列,表头wrap_text, 添加选项"""

	xlsxwriter_wb = Workbook(save_path)

	num_format = kwargs.get('num_format','0.00%')
	#通用表头格式
	background_color = rgb_convert_hex([220,230,241])
	border_color = rgb_convert_hex([149,179,215])
	header_format = xlsxwriter_wb.add_format({'font_name':'微软雅黑','bold':False,'font_size':11,
											  'bg_color':background_color })
	header_format.set_bottom()
	header_format.set_bottom_color(border_color)

	header_format.set_text_wrap()
	header_format.set_center_across()

	date_format =  xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':'yyyy/mm/dd'})
	pct_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':'0.00%'})
	text_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':'@'})
	normal_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11})

	df_list, sheet_name_list = df_sheet_check(df_list, sheet_name_list)

	for df,sheet_name in zip(df_list,sheet_name_list):
		xlsxwriter_ws = xlsxwriter_wb.add_worksheet(sheet_name)

		header_columns = list(df.columns)

		if not df.empty:
			pct_columns = [ x for x in header_columns if [y for y in pct_columns if y in x ]]
			#写入表头
			xlsxwriter_ws.write_row(row=0,col=0,data=header_columns,cell_format=header_format)

			for column_index,column in enumerate(header_columns):
				column_values = df[column].to_list()

				if column in pct_columns:  #如果属于百分比列 
					xlsxwriter_ws.write_column(row=1,col=column_index,data=column_values,cell_format=pct_format)
				#如果本列是日期类型,采用日期格式写入, 第二步应该到判断content_columns 还是 date? -- 以可以输入的参数为优先做判断
				elif np.issubdtype(df[column].dtype, np.datetime64) :
					column_format = date_format
				else:
					column_values = [ x if x == x else '' for x in column_values  ]
					column_format =  normal_format

				xlsxwriter_ws.write_column(row=1,col=column_index,data=column_values,cell_format=current_format)

			#表头加上选项
			xlsxwriter_ws.autofilter(0,0,len(df)-1,len(header_columns)-1)

			autofit_column_width(xlsxwriter_ws,header_columns,content_columns=content_columns,wrap_text=True)

			#冻结第一行窗口
			xlsxwriter_ws.freeze_panes(1,0)
		else: #如果没有数据，默认写入表头即可
			xlsxwriter_ws.write_row(row=0,col=0,data=header_columns,cell_format=header_format)

	save_xlsxwriter_wb(xlsxwriter_wb,save_path)

def replace_invalid_strs(string):
	invalid_strs = [ '\ud835' ]
	for i in invalid_strs:
		string = string.replace(i,'')
	return string 

def refresh_excel_calculations(file_path):
	if not os.path.isabs(file_path):
		file_path = os.path.join(os.getcwd(), file_path)

	xlapp = xw.App(visible=False)
	xlapp.display_alerts = False
	xlwb = xw.Book(file_path)
	xlwb.Visible = False
	xlapp.calculate()
	xlwb.save()
	xlapp.quit()

	del xlapp

# def refresh_excel_calculations(file_path):
# 	file_path = os.path.join(os.getcwd(), file_path )
# 	#检查输出文件夹 是否存在
# 	if not os.path.isfile(file_path):
# 		enter_exit("File not found: ",file_path)

# 	#打开文档，运行VBA, 如果有报错，everything 搜索gen_py文件 删除，让系统重新生成新的gen_py
# 	xlapp = client.gencache.EnsureDispatch('Excel.Application')   
# 	xlapp.Visible = 0  
# 	xlapp.DisplayAlerts = False 
# 	xlwb = xlapp.Workbooks.Open(file_path) 
# 	xlwb.Visible = 0
# 	xlapp.Calculate()
# 	xlwb.Close(True)
# 	xlapp.Quit

def write_format_columns(save_path,df_list,sheet_name_list=[],pct_columns=[],round_columns=[],content_columns=[],**kwargs):
	"""给定一个表格，用百分比写入包含指定文字的列,表头wrap_text, 添加选项,上面write_pct_columns方式不写是不希望每列str都变成text格式，
	   正常应该是常规格式
	   基本和前面的write_pct_columns相同，区别在于加上了整列如果都是str用text格式写入的方式"""
	xlsxwriter_wb = Workbook(save_path)

	min_column_width = kwargs.get('min_column_width', 12)

	round_list = kwargs.get('round_list',[ '标准差','平均','方差'])
	pct_list = kwargs.get('pct_list',['rate','百分比','占比','比例','比率'])

	#通用表头格式
	background_color = rgb_convert_hex([220,230,241])
	border_color = rgb_convert_hex([149,179,215])
	header_format = xlsxwriter_wb.add_format({'font_name':'微软雅黑','bold':False,'font_size':11,
											  'bg_color':background_color })
	header_format.set_bottom()
	header_format.set_bottom_color(border_color)

	header_format.set_text_wrap()
	header_format.set_center_across()

	date_format =  xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':'yyyy/mm/dd'})
	pct_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':'0.00%'})
	text_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':'@'})
	normal_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11})
	round_format = xlsxwriter_wb.add_format({'font_name':'calibri','font_size':11,'num_format':'0.00'})
	
	df_list, sheet_name_list = df_sheet_check(df_list, sheet_name_list)

	for df,sheet_name in zip(df_list,sheet_name_list):
		xlsxwriter_ws = xlsxwriter_wb.add_worksheet(sheet_name)

		#是否属于MultiIndex,如果是，需要确保写入完整--暂时不添加该功能
		# if isinstance(df.column, MultiIndex):

		header_columns = list(df.columns)

		if not df.empty:
			pct_columns = [ x for x in header_columns if [y for y in pct_columns if y in x ]]
			#写入表头
			xlsxwriter_ws.write_row(row=0,col=0,data=header_columns,cell_format=header_format)

			for column_index,column in enumerate(header_columns):
				column_values = df[column].to_list()
				if column in pct_columns or len([ x for x in pct_list if x in column]) > 0:  #如果属于百分比列 
					column_format = pct_format
				elif column in round_columns or len([ x for x in round_list if x in column]) > 0  : #如果属于需要round的列
					column_format = round_format
				#判断是否整列都属于str格式，采用text格式写入
				elif df[column].apply(type).eq(str).all():
					column_format = text_format
				#如果本列是日期类型,采用日期格式写入, 第二步应该到判断content_columns 还是 date? -- 以可以输入的参数为优先做判断
				elif np.issubdtype(df[column].dtype, np.datetime64)  :
					column_values = [ x if x == x else '' for x in column_values]
					column_format = date_format
				#是否整列都是float格式
				# elif df[column].apply(type).eq(float).all():
				# 	column_format = round_format
				else:
					column_values = [ x if x == x and type(x) == str else x for x in column_values  ]
					column_values = [ x if x == x else '' for x in column_values]
					column_format =  normal_format

				xlsxwriter_ws.write_column(row=1,col=column_index,data=column_values,cell_format=column_format)

			#表头加上选项
			xlsxwriter_ws.autofilter(0,0,len(df)-1,len(header_columns)-1)

			autofit_column_width(xlsxwriter_ws,header_columns,content_columns=content_columns,
											wrap_text=True, min_column_width=min_column_width)

			#冻结第一行窗口
			xlsxwriter_ws.freeze_panes(1,0)
		else: #如果没有数据，默认写入表头即可
			xlsxwriter_ws.write_row(row=0,col=0,data=header_columns,cell_format=header_format)

	save_xlsxwriter_wb(xlsxwriter_wb,save_path)

"""以下部分主要是openpyxl的方法"""
def rgb_convert_hex(rgb_list):
	#防止填入了超过255或者小于0的数值
	rgb_list = [ max(0,min(int(x),255)) for x in rgb_list]
	hex_value = '{0:02x}{1:02x}{2:02x}'.format(rgb_list[0],rgb_list[1],rgb_list[2]).upper()
	return hex_value


def get_column_row_num(worksheet):
	"""
	calculate the number of columns and rows from a worksheet 
	:param worksheet: openpyxl worksheet
	:return : column_letters, row_numbers 
	"""
	#计算一共有多少列
	column_num = len(list(worksheet['1']))
	column_letters = [get_column_letter(x).upper() for x in range(1,column_num+1)]

	#计算一共有多少行
	row_num = len(list(worksheet['A']))
	row_numbers = [ x for x in range(1,row_num+1) ]

	return column_letters,row_numbers

def format_custom_cells(worksheet,custom_cells_dict,custom_cell_color):
	"""
	处理特殊的单元格
	:param custom_cells_dict :默认key代表要修改的sheet_name,value是list类型，第一个元素是列，第二个元素的行
	:custom_cell_color: 指定单元格的字体颜色
	"""
	custom_cell_color = rgb_convert_hex(custom_cell_color)
	custom_font = Font(color=custom_cell_color)

	#需要调整的sheet_name
	#检查传入的sheet是否是需要修改的匹配的sheet_name
	for c_key in custom_cells_dict:
		if c_key == worksheet.title:
			#如果是这个需要修改的sheet 
			#第一个元素为列，第二个元素代表行
			c_column_list = custom_cells_dict[c_key][0]
			c_row_list = custom_cells_dict[c_key][1]

			#防止填入的数字有0
			if 0 in c_column_list:
				c_column_list = [x+1 for x in c_column_list]		
			if 0 in c_row_list:
				c_row_list = [x+1 for x in c_row_list]

			#循环修改格式
			for c_column in c_column_list:
				c_column_letter = get_column_letter(c_column)
				for c_row in c_row_list:
					c_target = c_column_letter + str(c_row)
					worksheet[c_target].font = custom_font

def format_table(worksheet,format_style,kwargs):
	"""
	传入worksheet进行统一格式调整后返回worksheet对象
	:param worksheet :openpyxl worksheet 
	:param format_style : 1/text or 2/data ,default is None
	:param wrap_text : the column letter you want to have auto line break, default is None
	:header_color : rgb value in list format, None is transparent, 'default' is light blue
	"""
	if kwargs == {}:
		kwargs = defaultdict()

	#表头颜色，最后一行字体是否加粗
	header_color = kwargs.get('header_color','default')
	last_row_bold = kwargs.get('last_row_bold',False)
	#第一行第一列的高度和宽度
	first_row_height = kwargs.get('first_row_height',27)
	first_column_width = kwargs.get('first_column_width',20)
	#主体部分宽和高
	main_part_width = kwargs.get('main_part_width',14)
	main_part_height = kwargs.get('main_part_height',14)
	#定制单元格
	custom_cells_dict = kwargs.get('custom_cells_dict',None)
	custom_cell_color = kwargs.get('custom_cell_color',[255,0,0]) #red 

	column_letters,row_numbers = get_column_row_num(worksheet)

	# =========== 通用的格式部分 ==========
	#第一行设置格式 _ 自动换行 _ 居中,是否旋转文字 等条件 
	row_1_alignment = Alignment(    horizontal = 'center',
									vertical = 'center',
									text_rotation=0,
									wrap_text=True,
									shrink_to_fit=False,
									indent=0  )
	#字体加粗
	row_1_font = Font(bold=True)

	#循环第一行格式
	for cell_1 in worksheet['1']:
		cell_1.alignment = row_1_alignment
		cell_1.font = row_1_font

	#表头颜色
	default_header_color = [189,215,238]  #浅蓝色
	
	if header_color == 'default':
		header_color = default_header_color

	if header_color != None:
		#把rgb的颜色格式转成aRGB hex value 
		header_color = rgb_convert_hex(header_color)
		#循环第一行背景颜色填充
		row_1_colorFill = PatternFill(start_color = header_color,
										end_color = header_color,
										fill_type ='solid')
		for cell_1 in worksheet['1']:
			cell_1.fill = row_1_colorFill

	#默认全部全部表格居左
	cell_alignment = Alignment(horizontal = 'left',vertical = 'center')

	for row in list(worksheet.rows)[1:]:
		for cell in row:
			cell.alignment = cell_alignment

	#循环调整行宽（默认行宽太窄）,以下代码容易出现错误 且不会提示哪里错，行对应是高，列对应是输入宽度
	#wrap_text很容易受别的alignment影响，在text格式下也不再进行调整
	for r in row_numbers:
		worksheet.row_dimensions[r].height = main_part_height
	#调整行高
	for c in column_letters:
		worksheet.column_dimensions[c].width = main_part_width

	#是否包含需要特殊格式的单元格
	if custom_cells_dict != None:
		format_custom_cells(worksheet,custom_cells_dict,custom_cell_color)

	# ========  1 TEXT .普通文字型表格 : 最后一列调整成特别的宽度============
	if str(format_style).lower() == 'text' or format_style == 1:
		worksheet.column_dimensions[column_letters[-1]].width = 80

		#wrap_text会被固定的行高限定住，如果想auto linebreak还是需要计算单元格里面的字体内容，通过内容和宽度得到应该调整的行高，
		#当前虽然没办法自动换行，但是还是需要添加这个条件，否则打开excel选择自动换行也不生效，还要选个行高自动调整
		wrap_text_alignment = Alignment(wrap_text=True)
		for cell in list(worksheet[column_letters[-1]])[1:]:
			cell.alignment = wrap_text_alignment

	# ============= 2 DATA 数据型表格 ==============
	if str(format_style).lower() == 'data' or format_style == 2:
		#第一行加高和第一列加宽
		worksheet.row_dimensions[1].height = first_row_height
		worksheet.column_dimensions['A'].width = first_column_width

		#循环第一列居左
		for cell_A in worksheet['A']:
			cell_A.alignment = Alignment(horizontal='left',vertical='center')

		#循环调整其他部分的全部列宽
		for c in column_letters[1:]:
			worksheet.column_dimensions[c].width = main_part_width
		#循环调整其他部分的全部行高
		for r in row_numbers[1:]:
			worksheet.row_dimensions[r].height = main_part_height

		#循环数字部分主体靠右展示
		for row in list(worksheet.rows)[1:]:
			for cell in list(row)[1:]:
				cell.alignment = Alignment(horizontal='right',vertical='center')

		#最后一行是否要加粗
		if last_row_bold == True:
			for cell in list(worksheet.rows)[-1]:
				cell.font = Font(bold=True)
				cell.border = Border(bottom=Side(border_style="thin", color="000000"))

	return worksheet 


def save_xlsxwriter_wb(xlsxwriter_wb,save_path):
	#检查是否存在save_path路径
	check_create_new_folder(save_path)
	close_tag = 0
	while close_tag <= 0:
		try:
			xlsxwriter_wb.close()
			close_tag += 1 
		except FileCreateError:
			input('\nFailed to write file!\n  Please Close "{}" Then Press Enter to Continue'.format(save_path))
	print('{0} Saved'.format(save_path))


def save_csv(df, save_path):
	check_create_new_folder(save_path)
	close_tag = 0
	while close_tag <= 0:
		try:
			df.to_csv(save_path,index=False)
			close_tag += 1 
		except PermissionError:
			input('\nFailed to write file!\n  Please Close "{}" Then Press Enter to Continue'.format(save_path))
	print('{0} Saved'.format(save_path))


def save_excel(df_list,save_path,sheet_name_list=None,format_style=None,**kwargs):
	"""
	通过openpyxl写入pandas表格，可以修改写入格式
	:param df_list : df list need to write 
	:param save_path : target path
	:param sheet_name_list : sheet_names coresponding to df_list
	:param format_style : default is None, bolded header;  text or 1  (change header format 
	 and column width only); 'data' or 2 adjust the header and the first column
	"""
	wb = openpyxl_workbook()

	#防止填入的不是列表格式
	if type(df_list) != list:
		df_list = [df_list]
	if type(sheet_name_list) != list:
		sheet_name_list = [sheet_name_list]
	
	if len(df_list) != len(sheet_name_list):
		print('写入的数据表数量和填入的Sheet名称数量不相等,将使用默认Sheet名称')
		
	if sheet_name_list == [None] :
		sheet_name_list= []
		for i in range(len(df_list)):
			sheet_name_list.append('Sheet {}'.format(i+1))
	
	for df, sheet_name in zip(df_list,sheet_name_list):
		if not df.empty:
			ws = wb.create_sheet(title=sheet_name)
			#写入表头
			column_list = df.columns.tolist()
			ws.append(column_list)
			counter = 0 
			for value in df.values:
				value = [ x.replace('\x01','') if type(x) == str  else x for x in value ]
				counter += 1 
				try:
					ws.append(list(value))
				except:
					print(f'本行数据出错：{sheet_name},{counter}')
					print(value)
			ws = format_table(ws,format_style=format_style,kwargs=kwargs)

	#删除默认的sheet
	del wb['Sheet']

	#保存全部文档
	check_close = 0  
	while check_close <= 0:
		try:
			wb.save(save_path)
			check_close += 1 
		except PermissionError:
			input('无法正常记录，请关闭{}后摁回车键继续'.format(save_path))

	print('{0} 已保存'.format(save_path))





import gc 
import re 
import sys  
import warnings 
import os 
import time  
from datetime import datetime 
import warnings   
import pandas as pd
import numpy as np
import hashlib
from collections import defaultdict,Counter

from .sequence_functions import list_diff_outer_join, lcs, filter_lcs
from .os_functions import *
from .df_functions import *
from .config_table import ConfigReader 
from .excel_functions import write_format_columns
from .regex_functions import replace_re_special, replace_punctuations
from .decorator_functions import *
from .data_handle_func import * 





import re  

country_map_dict={
"MY":"马来西亚",
"MM":"缅甸",
"TH":"泰国",
"IN":"印度",
"ID":"印尼",
"PH":"菲律宾",
"VN":"越南",
"CHN":"中国",
"PK":"巴基斯坦",
"KH":"柬埔寨",
"HK":"香港",
"LK":"斯里兰卡",
"RU":"俄罗斯",
"NP":"尼泊尔",
"BD":"孟加拉",
"TW":"台湾",
"SG":"新加坡",
"KZ":"哈萨克斯坦",
"MA":"摩洛哥",
"AU":"澳大利亚",
"AE":"阿联酋",
"OM":"阿曼",
"QA":"卡塔尔",
"KW":"科威特",
"YE":"也门",
"4C":"中东4C",
"LB":"黎巴嫩",
"JO":"约旦",
"NG":"尼日利亚",
"SA":"沙特阿拉伯",
"UA":"乌克兰",
"PL":"波兰",
"MO":"澳门",
"BT":"不丹",
"BN":"文莱",
"LA":"老挝",
"EG":"埃及",
"CH":"瑞士",
"KE":"肯尼亚"
}

engine_text = "mysql://root:00000000@localhost:3306/web_data?charset=utf8mb4"

# range_counter = 0 
# for i in range(0,50,10):
# 	range_counter += 1 
# 	print(i)
# 	if range_counter == 50 / 10 :
# 		if i != 50:
# 			print(i)


import gc 
import re 
import sys  
import warnings 
import os 
import time  
from datetime import datetime 
import warnings   
import pandas as pd
import numpy as np
import hashlib
from collections import defaultdict,Counter

from .sequence_functions import list_diff_outer_join, lcs, filter_lcs
from .os_functions import *
from .df_functions import *
from .config_table import ConfigReader 
from .excel_functions import write_format_columns
from .regex_functions import replace_re_special, replace_punctuations
from .decorator_functions import *
from .data_handle_func import * 

from pandas.core.indexes.multi import MultiIndex

class CsvSheetClass(object):
	def __init__(self, table_path):
		self.name = '{}'.format(table_path)
		self.visibility = 0

class Handler(object):

	def __init__(self, require_file_dir, input_dir,table_dict):
		#获取以下两个即可读取所有的原始数据和规则表
		#original data files 
		self.input_dir = input_dir
		#config table dict 
		self.table_dict = table_dict
		self.require_file_dir = require_file_dir

		# #从concat_data得到的结果记录输入数据的最大最小日期
		self.min_max_date_range = '' 

		#concat_data之后 保存一个原始表提供给后面的match做提取
		self.original_complete_header_df = None

	@catch_and_print
	def get_original2cn_dict(self, header_table_df, file_tag):
		"""
		将所有原始mapping成中文表头,按国家区分字典
		"""
		original2cn_dict_list = []
		original2cn_dict = defaultdict(str)
		fillna_dict = {}
		dtype_dict = {}

		if file_tag.lower() not in [ x.lower() for x in header_table_df.columns.get_level_values(0) ] :
			file_tag = 'Without file tag'

		header_column_index = header_table_df.columns.get_level_values(
			0) == file_tag.lower()

		header_table_df_c = header_table_df.iloc[:, header_column_index]

		header_table_first_three_c = header_table_df.loc[:, header_table_df.columns.get_level_values(0)[0]]

		# 同时获取填充的
		for row, last_three in zip(header_table_df_c.iterrows(), header_table_first_three_c.iterrows()):
			# 表头统一小写，换行符,空格全部去掉
			row_list = row[1].values
			last_three_list = last_three[1].values
			a_list = list(row_list)
			b_list = list(last_three_list)
			a_list = [str(x).lower().strip().replace('\n', '').replace('\xa0', '').replace(' ', '').replace('\t', '')
					  for x in a_list if x.strip() != '无' and x.strip().lower() != 'none' and x.strip() != '/' and x.strip() != '']

			if a_list:
				for x in a_list:
					original2cn_dict[x] = b_list[2]

			# 构建需要合并前填充的字典
			c_list = [x for x in a_list if split_colon(x)[0].lower().strip() == 'fillbeforeconcat' or split_colon(x)[0].strip() == '合并前填充']

			if c_list:
				for x in c_list:
					fillna_dict[b_list[2]] = split_colon(x)[1]

			if (b_list[1] != '默认' and b_list[1].lower() != 'default' and b_list[1] != '') and b_list[2] != '':
				dtype_dict.update({b_list[2]: b_list[1]})

		return original2cn_dict, fillna_dict, dtype_dict

	#合并读取的数据表格, 该函数需要输入table_dict因为需要读取到, complete_header_df, 和target_cn_columns
	@get_run_time
	def concat_data(self ):
		# 此函数读取放入的数据表，必须要运行
		for keys in self.table_dict.keys():
			if 'mapping' in keys.lower():
				mapping_key = keys
		try:
			header_table_df = self.table_dict[mapping_key]
		except KeyError:
			enter_exit('Cannot find mapping configuration sheet!')

		complete_header_df = self.table_dict['complete_header_df']
		target_cn_columns = self.table_dict['target_cn_columns']

		header_table_df = df_fillna_str(header_table_df)
		info_path_list = get_walk_abs_files(self.input_dir)

		# 检查是否有读取到各国的原始数据
		info_path_list = [x for x in info_path_list if '~$' not in x and (
			x[-5:].lower() == '.xlsx' or x[-4:].lower() in ['.xls', '.csv'])]

		if len(info_path_list) == 0:
			enter_exit(f'Cannot find any data file in folder "{self.input_dir}" !\n')

		success_sheet_df_list = []

		for table_path in info_path_list:
			table_p = Path(table_path)
			table_stem = table_p.stem
			table_suffix = table_p.suffix

			# 读取文件名的信息
			file_tag = table_stem.split('-')[0].split('_')[0].strip()

			# 获取这个文档的映射字典  将原始mapping成中文表头
			original2cn_dict, fillna_dict, dtype_dict = self.get_original2cn_dict(header_table_df, file_tag)

			if not original2cn_dict:
				enter_exit('"Data_processing_configuration" required mapping field "{}" not found !'.format(file_tag))

			# 如果是CSV文档
			is_csv = False
			is_xls_special = False
			if table_suffix == '.csv':
				is_csv = True
				csv_sheet_class = CsvSheetClass(table_stem)
				sheets_property_list = [csv_sheet_class]
			else:
				try:
					df_workbook = pd.ExcelFile(table_path)
					sheets_property_list = df_workbook.book.sheets()
					#试下能不能读取第一个sheet
					df_workbook.parse(str(sheets_property_list[0].name))
				except : #如果读取失败，尝试读取其他国家xls文档的格式
					is_xls_special = True 
					xls_sheet_class = CsvSheetClass(table_stem)
					sheets_property_list = [xls_sheet_class]

			# 过滤掉模板数据
			for sheets_property in sheets_property_list:
				sheet = sheets_property.name
				sheet_visibility = sheets_property.visibility
				if sheet_visibility == 0:  # 只读取可见的Sheet

					if is_csv:
						df_worksheet = read_csv_data(table_path)
						if df_worksheet.empty:
							continue
					elif is_xls_special: #这个格式的只读第一个sheet
						df_worksheet = read_xls_special(table_path)
						if df_worksheet.empty:
							continue
					else:
						df_worksheet = df_workbook.parse(str(sheet), na_values='')

					# 表头做小写等替换并且，通过字典rename,全部调整成去掉中间空格、去掉一切无意义符号的字段
					df_worksheet.columns = [str(x).lower().strip().replace('\n', '').replace('\xa0', '')
											.replace(' ', '').replace('\t', '')if x == x else x for x in 
											df_worksheet.columns]

					df_worksheet = dropping_not_mapping(df_worksheet, original2cn_dict, target_cn_columns)

					#mapping填入了 + 号
					df_worksheet = combine_multi_plus(df_worksheet, original2cn_dict)

					#mapping前检查是否有重复的字段，如果原表已经有别的字段映射成"机型"，那原表里面的"机型"字段属于要抛弃的字段
					df_work_sheet = drop_duplicated_columns_before_rename(df_worksheet, original2cn_dict)

					df_worksheet = df_worksheet.rename(original2cn_dict, axis=1)

					# 还必须要确认映射的字段没有重复，否则会影响到后面的数据列, 返回一个没有重复的字段列
					df_work_sheet = check_mapping_duplicates(df_worksheet, target_cn_columns, table_stem=table_stem)

					# 重命名之后，合并前需要填充默认值
					df_worksheet = fillna_with_dict(df_worksheet, fillna_dict)

					# 检查完重复映射之后 需要再定位一次需要的字段, 注意处理顺序
					df_worksheet = func_loc(df_worksheet, target_cn_columns)
					
					if not df_worksheet.empty:
						check_mapping_complete(df_worksheet, complete_header_df, original2cn_dict,file_tag=file_tag)

						#做一次字段格式处理，可以提示在哪个文档转错，但后面合并还是会把date转成object,所以需要再转一次
						complete_header_df = dtype_handle(complete_header_df, dtype_dict)
						# 记录成功表格
						success_sheet_df_list.append([table_stem, sheet, df_worksheet.shape[0]])
						#complete_header_df 是一个完整的表头，可以避免concat造成的表头混乱/缺失，
						#但合并会导致字段全变成object(CSV文档，xlsx的输入不受影响)
						complete_header_df = pd.concat([complete_header_df, df_worksheet], axis=0, sort=False, ignore_index=True)

						print(f'Getting from: "{table_stem}",Sheet:{sheet}, {df_worksheet.shape[0]} rows')

			success_sheet_df = pd.DataFrame(success_sheet_df_list, columns=['File Name', 'Source Sheet', 'Success reading records'])

		complete_header_df = dtype_handle(complete_header_df, dtype_dict)

		self.min_max_date_range = get_min_max_date(complete_header_df)
		self.original_complete_header_df = complete_header_df.copy()

		print(f'Data rows:{complete_header_df.shape[0]}')
		return complete_header_df, success_sheet_df


	 # 1.字段标准化: 是否考虑先做关联匹配以加速运行? -- 不可行，关联的国家字段需要先标准化才能用来统一关联匹配
	@get_run_time
	def standardize_columns(self, complete_header_df, standardize_config_df):
		# 将字段处理为标准表里的字段（模糊匹配）
		table_values = standardize_config_df.values
		not_standardize_df_list = []
		complete_header_df_sub_list = []
		partial_match_not_match_df_list = []

		row_counter = 0
		for row in table_values:
			row_counter += 1
			source_column = row[1]
			standard_table_name = row[2]
			standard_column = row[3]
			target_column = row[4]
			target_column_edit = row[5]
			order_columns = row[6]
			replace_dict = get_replace_dict(row[7])
			special_syn = [x.lower() for x in row[8].split('\n')]  #两边需要同时存在的字符
			filter_condition = row[9]
			# 标准化模式：简单模糊匹配 -- simple_lcs  严格模糊匹配 -- filter_lcs, 内存配置匹配--number_similarity
			standardize_mode = row[10]

			#决定最后的结果字段名称
			temp_column = ''
			if target_column_edit == source_column:
				temp_column = source_column
			else:
				temp_column = target_column_edit

			# 先把空值的数据剔除再做模糊匹配
			complete_header_df_notna = complete_header_df.loc[complete_header_df[source_column].isna() == False, :]
			complete_header_df_nan = complete_header_df.loc[complete_header_df[source_column].isna(), :]
			#过滤后全都是空值的话 直接继续下一行
			if complete_header_df_notna.empty:
				continue

			if standard_table_name != '' and standard_column.strip() != '':

				mode = get_standard_mode(standardize_mode)
				
				print(f'{row_counter}.Processing standardization for "{source_column}"')
				print(f'-- Referencing from table "{standard_table_name}" column "{target_column}",Mode:{standardize_mode}')
				standard_table_path = get_require_files(self.require_file_dir, [standard_table_name])[standard_table_name]
				# 统一将原始关联表转成str格式
				standard_table_df = read_config_table(standard_table_path)
				##读取出来的表 统一做删除重复字段处理, 如果模式是机型匹配模式
				standard_table_df = remove_duplicate_columns(standard_table_df)

				# 标准对照表排序,排序完之后删除重复项，似的做了排序后的结果取的是第一行上市日期最近的机型（后面循环的时候做duplicates删除）
				if order_columns != '':
					standard_table_df = process_sort_order(standard_table_df, [standard_column], order_columns)

				# 需要对标准表做重复删除，类似字段匹配, 但不相同
				# 标准化前的第一层简单过滤
				filter_condition_2_columns_tag, filter_left_column, filter_right_column = False, '', ''

				if filter_condition != '':
					filter_condition_2_columns_tag, filter_left_column, filter_right_column = \
						get_filter_condition_standardize_tag(filter_condition)

				if filter_condition != '' and filter_condition_2_columns_tag == False:
					try:
						standard_table_df = standard_table_df.query(filter_condition)
					except:
						enter_exit(f'Standardization: Failed to compile condition: {filter_condition}')

				#全部转成半角符号
				standard_table_df[standard_column] = standard_table_df[standard_column].apply(
															lambda x : normalize_punctuations(x) if type(x) == str else x ) 
				complete_header_df_notna[source_column] = complete_header_df_notna[source_column].apply(
															lambda x : normalize_punctuations(x) if type(x) == str else x)
				# 标准化前的第二层检查过滤(过滤条件涉及两张表的字段相等(不同国家的机型匹配))
				# 如果存在另外一种过滤方式--左右表的字段相等, 需要循环条件 进行模糊匹配
				if filter_condition_2_columns_tag:
					# 模糊关联匹配之前，必须做去重，防止笛卡尔积现象（模糊匹配防止获取的不是排序第一的数据）
					find_lack_columns( standard_table_df, [standard_column, target_column, filter_left_column])
					# 需要保留一个模糊匹配需要获取的 target_column
					standard_table_df_x = standard_table_df.loc[:, [standard_column, target_column, filter_left_column]]\
						.drop_duplicates(subset=[standard_column, filter_left_column], keep='first')

					# 循环获取模糊匹配结果
					if filter_right_column not in complete_header_df_notna.columns:
						enter_exit(f'Standardization Error: Cannot find column:"{filter_right_column}"')
					#记录每个的情况
					complete_header_df_sub_list = []
					for u in complete_header_df_notna[filter_right_column].unique() :
						temp_standard_df = standard_table_df_x\
								.loc[standard_table_df_x[filter_left_column] == u, [standard_column, target_column]]
						standard_dict = temp_standard_df.to_dict()

						print(f'Standardizing: "{u}"--"{source_column}"')

						standard_dict = {x: y for x, y in zip(
							standard_dict[standard_column].values(), standard_dict[target_column].values())}

						complete_header_df_sub = complete_header_df_notna.loc[
							complete_header_df_notna[filter_right_column] == u, :]

						complete_header_df_sub[temp_column] = complete_header_df_sub[source_column].fillna(value='').astype(str)

						complete_header_df_sub[temp_column] = complete_header_df_sub[temp_column].apply(
								lambda x: standardize_column_func(
								x, standard_dict, special_syn, replace_dict, ignore_punctuation=True, mode=mode))

						complete_header_df_sub_list.append(complete_header_df_sub)

					if len(complete_header_df_sub_list) == 1:
						complete_header_df_notna = complete_header_df_sub_list[0]
					elif len(complete_header_df_sub_list) >= 2 :
						complete_header_df_notna = pd.concat(complete_header_df_sub_list, axis=0, ignore_index=True)
					else: #complete_header_df_sub都没有,生成一列空白的结果
						complete_header_df_notna[temp_column] = ''

				# 如果是普通的df quiry过滤方式
				else:
					lack_column_list = find_lack_columns(
									standard_table_df, [standard_column, target_column],'Standardization reference table')
					# 提取出标准列表
					standard_dict = standard_table_df.loc[:, [standard_column, target_column]].to_dict()
					standard_dict = {x: y for x, y in zip(
						standard_dict[standard_column].values(), standard_dict[target_column].values())}

					complete_header_df_notna[temp_column] = complete_header_df_notna[source_column]\
							.apply(lambda x: standardize_column_func(x, standard_dict, special_syn, replace_dict,
													ignore_punctuation=True, mode=mode) if type(x) == str else x)

				# 空的和处理过的非空数据记得合并
				complete_header_df = pd.concat([complete_header_df_notna, complete_header_df_nan], axis=0, ignore_index=True)

				# 需要记录这两个字段分别有哪些记录匹配不上
				partial_match_not_match_df = get_partial_not_match(complete_header_df_notna,row_counter,source_column,
																	standard_table_name, standard_column,
																	filter_condition, target_column_edit,  filter_left_column)

				partial_match_not_match_df_list.append(partial_match_not_match_df)

		if partial_match_not_match_df_list:
			partial_match_not_match_df = pd.concat(partial_match_not_match_df_list, axis=0, ignore_index=True)

		# 模糊匹配的因为涉及"获取结果的字段名称"可能会修改成原始字段名称，判断复杂, 故不做记录
	
		print(f'Data rows:{complete_header_df.shape[0]}')
		return complete_header_df, partial_match_not_match_df

	# 2.字段拆分
	@get_run_time
	def split_columns(self, complete_header_df, split_table_df):

		#拆分字段,并做堆叠, 获取的配置格式 {'拆分字段':[ 标准词组列表]} 
		split_table_df = df_fillna_str(split_table_df)
		split_table_values = split_table_df.values

		split_config_dict = { }
		for row in split_table_values:
			split_column = row[1]
			split_table_name = row[2]
			split_table_column = row[3]
			filter_condition = row[4]
			split_mode = row[5]
			split_symbol = row[6]

			if split_table_column != '' and split_table_name.strip() != '' and split_column.strip() != '' :

				split_file_dict = get_require_files(self.require_file_dir,[split_table_name])
				split_file_path = split_file_dict[split_table_name] 
				split_standard_df = read_config_table(split_file_path)
				split_standard_df.columns = [ x.strip() if type(x) == str else x for x in split_standard_df.columns  ]

				if filter_condition:
					split_standard_df = df_query(split_standard_df, filter_condition)
				if split_table_column not in split_standard_df.columns:
					enter_exit(f'Failed to find column "{split_table_column}" in table "{split_table_name}" ')
				else:
					split_standard_list = [ x for x in split_standard_df[split_table_column].tolist() if x == x and x != '' ] 
					split_standard_list = sorted(split_standard_list, key= len , reverse=True)
				split_config_dict.update({split_column: [split_standard_list, split_symbol ]})
			else:
				print(f'Standard column used to split is empty, "{split_symbol}" will be the symbol used to split')
				split_config_dict.update({split_column:[[], split_symbol] })

		complete_header_df = split_column_by_words(complete_header_df, split_config_dict, mode=split_mode)

		print(f'Data rows:{complete_header_df.shape[0]}')
		return complete_header_df

	# 3.字段匹配
	@get_run_time
	def match_columns(self, complete_header_df, match_table_df):
		table_values = df_fillna_str(match_table_df).values

		not_match_df_list = []
		for row in table_values:
			source_columns = split_colon(row[1])  # 目的表字段
			join_table_name = row[2]  # 关联表
			join_columns = split_colon(row[3])    # 关联字段
			target_columns = split_colon(row[4])  # 想获取的目标字段
			target_columns_edit = split_colon(row[5])  # 结果字段名称
			sort_order = row[6]  # 匹配前的去重排序
			replace_dict = get_replace_dict(row[7]) 
			filter_condition = row[8]  # 关联表过滤条件
			match_mode = row[9]  # 匹配模式：1.case insensitive 2.case sensitive
			# For columns with spaces in their name, you can use backtick quoting.
			
			#如果目标字段和重命名的目标字段 长度不等，直接全部改成所需要的字段名称
			if len(target_columns_edit) != len(target_columns) :
				target_columns_edit = target_columns 

			if source_columns !='' and len(join_columns) > 0 and len(target_columns) > 0  :
				#如果两边的字段数量不一致
				if len(source_columns) != len(join_columns):
					min_num = min([len(source_columns), len(join_column)])
					source_columns = source_columns[:min_num+1]
					join_columns = join_columns[:min_num+1]

				#如果填了匹配表，开始读取
				if join_table_name != '':
					join_table_path = get_require_files(self.require_file_dir, [join_table_name])[join_table_name]
					join_table_df = read_config_table(join_table_path)
				else: #如果没有填匹配表，则使用原始输入表格进行匹配
					join_table_df = self.original_complete_header_df

				# 检查两张表 有没有字段缺失
				find_lack_columns(complete_header_df, source_columns,'Matching, complete_header_df')
				find_lack_columns(join_table_df, set(join_columns + target_columns), 'Matching, join_table')

				join_table_df = process_join_table(join_table_df=join_table_df, join_columns=join_columns, 
												   target_columns=target_columns, filter_condition = filter_condition, 
												   sort_order = sort_order, join_table_name = join_table_name)

				complete_header_df = process_match_complete_table(complete_header_df= complete_header_df, 
																  source_columns = source_columns ,target_columns = target_columns, 
																  join_columns = join_columns, join_table_name = join_table_name)

				#重命名进行匹配，必须要做的，无论是否用的外部表匹配
				for i in range(len(source_columns)):
					join_table_df[source_columns[i]] = join_table_df[join_columns[i]]

				# 防止只匹配表只有一列，并且只想获取该列结果
				join_table_df, only_one_match_column = check_only_one_match_column(join_table_df, join_columns, target_columns )

				# 匹配默认忽略大小写
				if type(match_mode) == str and match_mode.lower().replace('-', '') == 'case sensitive':
					complete_header_df = pd.merge(complete_header_df, join_table_df, 'left', on=source_columns)
				else:
					complete_header_df = merge_case_insensitive(complete_header_df, join_table_df, 'left', on=source_columns)
				if only_one_match_column:
					complete_header_df = complete_header_df.rename({'additional_temp': target_columns_edit[0]}, axis=1)

				#重命名，直接列表重新赋值即可
				if target_columns_edit != target_columns:
					for t1, t2 in zip(target_columns, target_columns_edit):
						complete_header_df[t2] = complete_header_df[t1]

				# 记录无法匹配到结果的数据，第一个target_columns如果为空，判断为无法匹配
				not_match_df = complete_header_df.loc[complete_header_df[target_columns_edit[0]].isna(), source_columns]\
				.drop_duplicates()

				not_match_df['Content'] = not_match_df[source_columns].agg('+'.join, axis=1)

				not_match_df['Matching field'] = '+'.join(source_columns) + ' Match ' + '+'.join(join_columns)

				#转换表头顺序
				shift_order_list = list(not_match_df.columns)
				not_match_df = not_match_df.loc[:,['Matching field','Content']]

				not_match_df_list.append(not_match_df)

		if join_table_name != '':
			not_match_df = pd.concat(not_match_df_list, axis=0, ignore_index=True)

			not_match_df = not_match_df.drop_duplicates()
		else:
			not_match_df = pd.DataFrame([])

		print(f'Data rows:{complete_header_df.shape[0]}')
		return complete_header_df, not_match_df

	# 4. 字段去重, 顺便生成MD5
	@get_run_time
	def drop_duplicate_data(self, complete_header_df, drop_duplicates_table_df):
		table_values = drop_duplicates_table_df.values
		drop_subset_list = []
		md5_unique_column_list = []
		for row in table_values:
			column = row[0]
			if_drop_duplicate = row[1]
			if_gen_md5 = row[2]

			if column.strip() != '' and column in complete_header_df.columns:
				if if_drop_duplicate.lower().strip() in ['是', 'yes', 'y']:
					drop_subset_list.append(column)
				if if_gen_md5.lower().strip() in ['是', 'yes', 'y']:
					md5_unique_column_list.append(column)
			else:
				enter_exit(
					f'Deduplicate reference table error: Column "{column}" not found! ')

		if drop_subset_list:
			complete_header_df = complete_header_df.drop_duplicates(
				subset=drop_subset_list)

		if md5_unique_column_list:
			complete_header_df = column_gen_md5(
				complete_header_df, md5_unique_column_list, 'MD5')

		print(f'Data rows:{complete_header_df.shape[0]}')
		return complete_header_df

	# 5.填充排序
	@get_run_time
	def fill_and_sort_columns(self, complete_header_df, fill_and_sort_table):
		
		table_values = fill_and_sort_table.values
		
		source_columns = [ row[0] for row in table_values]
		complete_header_df_new = pd.DataFrame([])

		seen_columns = set()
		sort_column_list = []
		sort_column_order_list = []
		dtype_dict = {}

		for row in table_values:
			input_column = row[0]
			output_column = row[1]
			# 输出的类型 日期类型--不包含时分秒 normalize().date 时间类型-包含时分秒-to_datetime, 和其他pandas支持的数据类型
			output_dtype = row[2]
			# 先做内容替换，再做空值填充
			replace_value_str = row[3]
			sort_value_str = row[4]
			fillna_value_str = row[5]

			if input_column != '' and output_column != '':
				if input_column not in complete_header_df.columns:
					if_skip = input(
						f'"{input_column}" not found in result table, thus not able to be mapped to the output result, continue?(Enter to continue)')
					if if_skip == '' or if_skip.lower() == 'yes' or if_skip.lower() == 'y':
						continue
					else:
						enter_exit('')
				#先将旧数据赋值到新数据 , 如果是重复填入的则保持用complete_header_df_new的数据处理即可
				if input_column not in seen_columns:
					complete_header_df_new[output_column] = complete_header_df[input_column].fillna('')
					seen_columns.add(input_column)
					
				if input_column != '' and output_column != '':
					if replace_value_str != '':
						complete_header_df_new = replace_value_func(
							complete_header_df_new, replace_value_str, input_column, output_column)

					# 过滤： 以下全部沿用上面获得的complete_header_df_new和他的所有output_column
					if fillna_value_str != '':
						for fillna_value_str_x in split_colon(fillna_value_str):
							complete_header_df_new = fillna_value_func(
								complete_header_df, complete_header_df_new, fillna_value_str_x, output_column)
					# 排序：统一加到列表，统一排序, 通过正-顺序，负数-倒序，数字代表是优先第几sort的顺序
					if sort_value_str.strip() != '':
						try:
							sort_value_number = int(float(sort_value_str))
							sort_column_list.append(output_column)
							sort_column_order_list.append(sort_value_number)
						except:
							pass
					# 输出类型：
					if output_dtype != '':
						dtype_dict.update({output_column: output_dtype})

		complete_header_df_new = dtype_handle(
			complete_header_df_new, dtype_dict, output=True)
			
		if sort_column_list:
			complete_header_df_new = sort_value_func(
				complete_header_df_new, sort_column_list, sort_column_order_list)

		print(f'Data rows:{complete_header_df_new.shape[0]}')

		return complete_header_df_new

	# 6.条件过滤
	@get_run_time
	def filter_columns(self, complete_header_df, filter_condition_table):

		table_values = filter_condition_table.values
		for row in table_values:
			condition = row[0].strip()
			complete_header_df = df_query(complete_header_df, condition)

		print(f'Data rows:{complete_header_df.shape[0]}')
		return complete_header_df

	@get_run_time
	def regex_extraction(self, complete_header_df, regex_extraction_table):

		table_values = regex_extraction_table.values

		for row in table_values:
			source_column = row[1]
			output_column = row[2]
			extract_pattten = row[3]

			find_lack_columns(complete_header_df, [source_column])

			if source_column != '' and extract_pattten != '':
				complete_header_df[output_column] = ''
				complete_header_df[output_column] = complete_header_df[source_column]\
				.apply(lambda x: '\n'.join(re.findall(extract_pattten,x,flags=re.I)).strip() if type(x) == str else '')

		return complete_header_df


	#***************以下是数据统计的函数***************************#

	#如果配置表有时间的处理的部分，传入需要计算的时间周期
	def calc_time(self, data_df,time_process_df):

		if not time_process_df.empty and not data_df.empty:

			for row in time_process_df.values:
				time_column = row[1]
				enddate = row[2]
				original_period = row[3]
				period = row[4].strip().lower()

				#获取时间范围对应的统计截止时间, 如果输入的是原始数据表格的字段，将该字段转成时间格式
				if enddate in data_df.columns:
					data_df = normalize_column_dates(data_df,enddate)
					enddate = data_df[enddate]
				else:
					enddate = normalize_dates_single(enddate)

				#获取需要统计的时间区间
				re_result = re.search('^([a-zA-Z\u4e00-\u9fa5]{1,}).?(\d{1,})$',original_period)

				#确保输入的时间区间格式
				try:
					period = re_result.group(1)
					period_number = re_result.group(2).strip()
				except:
					enter_exit('Time range input error: Make sure the input time range is like day_1, day_2, week_1, etc')

				if time_column not in data_df.columns:
					enter_exit(f'Time process error: Missing column:{time_column}')

				#将原始时间字段转成时间格式  
				data_df = normalize_column_dates(data_df, time_column)

				#用截止时间来计算时间不同的维度，day, week , month, year
				#使用np.timedelta64来计算 需要将日期转成pandas to_datetime的日期格式
				if period in ['day','days','天','日','d']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'D')
				elif period in [ 'week','weeks', '周','星期','w']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'W')
				elif period in [ 'month','months', '月','月数','m']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'M')
				elif period in [ 'year','years', '年','年数','y']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'Y')
				elif period in [ 'hour','hours', '时','小时','h']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'h')
				elif period in [ 'minute','minutes', '分','分钟','m']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'m')
				elif period in [ 'sales week', 'sales weeks','实销周','实销周数']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'W')
				elif period in [ 'sales month','sales months','实销月','实销月数']:
					data_df[original_period] = (enddate - data_df[time_column])/np.timedelta64(period_number,'M')
				else:
					enter_exit(f'Unknown time period:{original_period}')

				#再检查一次时间是否完整
				data_df_empty_date = data_df.loc[data_df[time_column].isna(),:]
				if not data_df_empty_date.empty:
					enter_exit('Time process error: column "{}" has {} empty datetime rows after combining all the data files!'\
						.format(time_column,data_df_empty_date.shape[0]))
				#往上取整，统一转成int格式
				data_df[original_period] = data_df[original_period].apply(np.ceil).astype(int)

				#将结果转成True和False,因为后面统计不能用数字结果去做分组, 结果为1的就是属于填入的时间区间
				# data_df[original_period] = data_df[original_period].apply(lambda x: True if x==1 else False)
		else:
			return data_df 
		return data_df

	@catch_and_print
	def process_statistic_groups(self, data_df, statistic_groups_df):

		#防止读取的透视字段是合并的单元格
		group_column = list(statistic_groups_df.columns)[1]
		statistic_groups_df[group_column] = statistic_groups_df[group_column].replace('',float('Nan')).fillna(method='ffill')

		values = statistic_groups_df.values

		groups = sorted(list(set([ x for x in split_colon(values[0][3]) if x !=''])))
		# #防止有过滤条件的聚合之后 部分主键丢失导致无法做skipna的计算功能
		agg_df_sub = generate_complete_index(data_df,groups)

		temp_df_list = [ agg_df_sub ]

		for row in values:
			filter_condition = row[1]
			drop_duplicates_condition = row[2]
			groups_str = row[3]
			groups = sorted(list(set([ x for x in split_colon(row[3]) if x !=''])))
			value_column = row[4]
			agg_func = row[5].strip()
			result_name = row[6]

			if value_column :
				#检查是否少字段
				lack_column_list = find_lack_columns(data_df, groups, 'Grouping error')
				#过滤和去重
				data_df_sub = df_query(data_df,filter_condition)
				data_df_sub = process_duplicates(data_df_sub, drop_duplicates_condition)

				if 'word' in  agg_func: #如果为空 返回空字典
					if data_df_sub.empty: 
						agg_df_sub = create_group_empty(data_df, groups, value_column, fillna=r'{}')
					else:
						require_file_dict = get_require_files(self.require_file_dir, ['keyword','stop_word'])
						keyword_dict, keyword_format_dict = get_keyword_dict(require_file_dict['keyword'])
						stopword_dict, stopword_format_dict = get_keyword_dict(require_file_dict['stop_word'])

						keyword_list = convert_key2list(keyword_dict)
						stopword_list = convert_key2list(stopword_dict) 
						#词频统计需要传入关键词字典
						agg_df_sub = group_basic_agg(data_df_sub, groups,  agg_func, 
											value_column,keyword_list,stopword_list,group_index=True)
				else:
					if data_df_sub.empty:
						agg_df_sub = create_group_empty(data_df, groups, value_column, fillna=0 )
					else:
						agg_df_sub = group_basic_agg(data_df_sub, groups,  agg_func, value_column,group_index=True)

				agg_df_sub = agg_df_sub.rename({value_column:result_name},axis=1)

				temp_df_list.append(agg_df_sub)

		if temp_df_list :
			agg_df = pd.concat(temp_df_list, axis=1)

		agg_df = agg_df.fillna(0).reset_index()

		return agg_df

	@catch_and_print
	def process_calculations(self, data_df, calculation_df):
		counter = 0 
		for row in calculation_df.values:
			counter += 1 
			calc_content = row[1].strip()
			result_name = row[2]
			if calc_content != '':
				try:
					data_df[result_name] = data_df.eval(calc_content)
				except :
					enter_exit(f' Calculation Error on row {counter}, \n Content:{calc_content}')

		return data_df

	@catch_and_print
	def pivot_table(self, data_df, pivot_table_df):
		#只读取第一行
		row_counter = 0  
		for row in pivot_table_df.values:
			row_counter += 1 
			if all([ x == x and x != '' for x in row[:-1] ]) :
				index = [ x for x in split_colon(row[0]) if x != '' ] 
				columns =  [ x for x in split_colon(row[1]) if x != '' ] 
				values =  [ x for x in split_colon(row[2]) if x != '' ] 
				#检查所有字段
				find_lack_columns(data_df, index + columns + values)
				data_df = pd.pivot_table(data_df,index=index, columns=columns, values=values)

				#如果得到的结果有多个表头，只保留最后一行表头
				if isinstance(data_df.columns, MultiIndex):
					try: # index error will occur if df is empty
						level_elem = len(data_df.columns[0])
						drop_len = [ x for x in range(level_elem) ][:-1] 
					except:
						drop_len = [0]
					data_df.columns = data_df.columns.droplevel(level=drop_len)

				data_df = data_df.reset_index(drop=False)

		return data_df 


import gc 
import re 
import sys  
import warnings 
import os 
import time  
from datetime import datetime 
import warnings  
import numpy as np 
import pandas as pd
import hashlib
from xlsxwriter import Workbook 
from collections import defaultdict,Counter

from .os_functions import get_walk_files,get_walk_abs_files,\
check_require_files,check_create_new_folder,get_require_files,enter_exit
from .df_functions import normalize_multi_header,copy_seperate_header_columns,\
check_abnormal_dates, stack_list_column, df_fillna_str
from .excel_functions import write_pct_columns


warnings.filterwarnings('ignore')
warnings.simplefilter(action='ignore', category=FutureWarning)

from pandas.core.index import MultiIndex

class ConfigReader(object):

	def __init__(self,config_file_dir, config_table_name, config_list,*args, **kwargs):

		self.config_file_dir = config_file_dir
		self.config_table_name = config_table_name
		self.config_list = config_list 

		self.require_file_dir = kwargs.get('require_file_dir','.\\require_tables')
		self.data_file_dir = kwargs.get('data_file_dir',r"..\\data_files")

	def get_header_table(self,header_table_df):
		#过滤全都是空的行
		header_table_df = header_table_df.dropna(how='all',axis=0)
		header_table_df = df_fillna_str(header_table_df)
		header_table_df =  normalize_multi_header(header_table_df)

		return header_table_df

	def get_complete_header_df(self, header_table_df):
		#保留的标准表头数量 以第一列的序号为准
		header_table_columns = header_table_df.columns
		standard_column = header_table_df[header_table_columns[0]].fillna('').tolist()

		standard_column = [x for x in standard_column if x != '']

		for s in standard_column[::-1] :
			if s == '':
				standard_column.pop(-1)
			else:
				break

		target_column_num = len(standard_column)
		target_cn_columns = header_table_df[header_table_columns[2]][:target_column_num].tolist()

		complete_header_df = pd.DataFrame(data= [],columns=target_cn_columns)

		return complete_header_df,target_cn_columns 

	def get_config_tables(self, if_walk_path = True ):

		require_file_dict = get_require_files(self.config_file_dir, self.config_table_name,if_walk_path=if_walk_path)

		header_table_path = require_file_dict[self.config_table_name]
		
		df_workbook = pd.ExcelFile(header_table_path)

		sheet_property_list = df_workbook.book.sheets()

		table_dict = { }
		for sheet_property in sheet_property_list:
			sheet = sheet_property.name

			sheet_visibility = sheet_property.visibility

			if sheet_visibility == 0 :  #只读取可见的sheet
				for config in self.config_list:
					if config in sheet.lower().strip():
						#需要特殊处理的合并表
						if 'mapping' in config:
							table = df_workbook.parse(sheet, header = [0, 1])
							table = self.get_header_table(table)
							complete_header_df, target_cn_columns = self.get_complete_header_df(table)
						else:
							table = df_workbook.parse(sheet, header = 0 )

						if not table.empty:
							table = df_fillna_str(table)
							table_dict.update({sheet:table})

		table_dict.update({ 'complete_header_df' :complete_header_df,
							'target_cn_columns':target_cn_columns })

		return table_dict


if __name__ == '__main__':

	config_list = [ 'mapping',
					'standardization',
					'split',
					'match',
					'deduplication',
					'fill&sort',
					'filter',
					'extraction']


	config_list =[ 'mapping',
				   'time process',
				   'statistic groups',
				   'calculations',
				   'fill&sort']

	table_reader = ConfigReader(config_file_dir= '.\\',config_list=config_list,config_table_name= 'config',)

	table_dict = table_reader.get_config_tables(if_walk_path=False)

	df = table_dict['time process']

	print(df)



import pandas as pd 
from collections import defaultdict 
import datetime 
from xlrd import xldate_as_datetime
import os 
import sys
import json
from openpyxl import load_workbook
from .os_functions import last_day_of_month,enter_exit, generate_md5
from .regex_functions import replace_re_special, strQ2B, strB2Q,symbol_to_spaces, normalize_punctuations
from .nlp_functions import get_keyword_dict, get_word_freq_dict, convert_key2list, process_text_eng
from .excel_functions import write_format_columns
from .func_classes import DfDict
import gc 
import re 
import warnings 
import traceback 
import logging 
from pandas.errors import OutOfBoundsDatetime
import swifter
from flashtext import KeywordProcessor 

warnings.filterwarnings('ignore')


def read_config_table(file_path, dtype=str):
	df = pd.DataFrame([])
	#读取数据文件，只读取第一个肉眼可见的sheet
	if '.xlsx' == file_path[-5:]:
		df_workbook = pd.ExcelFile(file_path)
		sheets_property_list = df_workbook.book.sheets()
		for sheet_property in sheets_property_list:
			if sheet_property.visibility  == 0:
				df = df_workbook.parse(sheet_property.name, dtype=str)
				break 
	else:
		try:
			df = pd.read_excel(file_path)
		except:
			df = pd.read_html(file_path,header=0)

	if df.empty:
		enter_exit(f'Cannot read any visible table in "{file_path}"')
	return df 

def read_data_file(file_path):
	df = pd.DataFrame([])
	#读取数据文件，只读取第一个sheet
	if '.csv' == file_path[-4:]:
		df = pd.read_csv(file_path)
	elif '.xlsx' == file_path[-5:]:
		df = pd.read_excel(file_path)
	else:
		try:
			df = pd.read_excel(file_path)
		except:
			df = pd.read_html(file_path,header=0)
	return df 


#打开EXCEL文档前，激活一遍所有的sheet 防止包含有公式的单元格 读取到的数据不是最新的，比如=today()不刷新一遍，
#pandas读取到的永远是EXCEL文档保存的当天
def generate_complete_index(df, group_columns):
	#通过需要做成索引的字段，补充出一份完整的索引
	number_column = ''

	original_group_columns = group_columns[:]
	#获取到数字列的最大即可，文字列 不用交叉生成新数据
	column_min = 0 
	column_max = 0 
	num_unique_values = [ ]

	total_df_list = [ ]
	for c in group_columns:
		row_list = [ ]
		try: #是否是数字列， 如果是获取最大最小值, 通常不会拿float当成索引
			df[c] = df[c].astype(float)
			#转成float之后 检查是否全部字段都是空了，如果是空，直接跳过
			df_not_na = df.loc[df[c].isna()==False,:]
			c_min = int(df[c].fillna(0).min(skipna=True))
			c_max = int(df[c].fillna(0).max(skipna=True))
			if c_max > column_max:
				column_max = c_max
				number_column = c
			if c_min < column_min:
				column_min = c_min

			if column_max < 5: #最小保证要是5以上(week_1 报错)
				num_unique_values = [ x for x in range(0,11)]
			elif column_max >= 5 and column_max <= 15: #临时限定补充的部分不超过12，通常计算0不会超过这个(天/周维度)，否则计算的数据量太大
				num_unique_values = [ x for x in range(column_min, column_max + 1 )]
			else:
				num_unique_values = [ x for x in range(0,16) ]

		except: #如果不是数字列，获取所有唯一值
			pass 

	#如果存在数字列,生成完整的df 
	if num_unique_values:
		try:
			group_columns.remove(number_column)
		except KeyError:
			enter_exit('Error when trying to generate complete date index, please check datetime column value in config files!')
		df_temp = df.loc[:,group_columns]
		df_temp = df_temp.drop_duplicates()

		for i in num_unique_values:
			df_temp[number_column] = i
			total_df_list.append(df_temp.copy())

		df_temp_total = pd.concat(total_df_list,axis=0,ignore_index=True)
	else:
		find_lack_columns(df,group_columns)
		df_temp_total = df.loc[:,group_columns]
		df_temp_total = df_temp_total.drop_duplicates()

	df_temp_total = df_temp_total.set_index(original_group_columns)

	return df_temp_total


#创建一个完整的主键
def set_index_drop_all(df, index_columns):
	df_copy = df.copy()
	#指定特定的索引，清空所有字段数据（为了可能的有过滤的运算，做左关联的完整匹配）
	drop_columns = set(df_copy.columns) - set(index_columns)
	df_copy = df_copy.drop_duplicates(subset=list(index_columns))
	df_copy = df_copy.set_index(list(index_columns))
	df_copy['temp'] = True
	df_copy = df_copy.drop(drop_columns,axis=1)

	return df_copy

def df_query(df, condition):
	if type(condition) == str and condition.strip() != "":
		try:
			df = df.query(condition,engine='python')
		except:
			write_format_columns('Filter condition-Result when error.xlsx',df,'content')
			enter_exit(f'Unable to compile the following filter condition：\n"{condition}"')
			
	return df 

def find_lack_columns(df, require_columns, error_func=''):

	lack_column_list = [ ]
	for r in require_columns:
		if r not in df.columns:
			lack_column_list.append(r)

	if lack_column_list:
		error_msg = ','.join(lack_column_list)
		enter_exit(f"{error_func} Error - Missing columns:{error_msg}")

	return lack_column_list


def column_gen_md5(df,unique_id_columns,target_column='MD5'):
	#几列合并 sum(多个字段) 
	df[target_column] = df.loc[:,unique_id_columns].fillna('').astype(str).sum(1).apply(lambda x : generate_md5(x))
	return df 

def fillna_with_dict(df,fillna_dict):
	for k,  v in fillna_dict.items():
		if k in df.columns:
			df[k] = df[k].fillna(value=v) 
		else:
			df[k] = v 
	return df


def df_fillna_str(df):
	df = df.fillna(value='').astype(str).apply(lambda x: x.str.strip())
	return df  

#定位目标字段
def func_loc(df,used_cols):

	new_used_cols = [ ]
	#防止重复定位同一个字段
	seen_column = [ ]
	for u in used_cols:
		if u not in seen_column:
			new_used_cols.append(u)

		if u not in df.columns:
			df[u] = ''

	df = df.loc[:,new_used_cols]

	return df

def merge_case_insensitive(df_1,df_2,how, on):
	#忽略大小写和忽略所有全角半角符号匹配
	temp_list = [ ]
	for i, o in enumerate(on) :
		temp_name = f'lower_case_{i+1}' 
		df_1[temp_name] = ''
		df_2[temp_name] = ''
		#需要统一转成str格式去做匹配
		df_1[temp_name] = df_1[o].fillna('').astype(str).apply(lambda x: normalize_punctuations(x).lower().strip())
		df_2[temp_name] = df_2[o].fillna('').astype(str).apply(lambda x: normalize_punctuations(x).lower().strip())

		temp_list.append(temp_name)

	if how == 'right':
		df_1 = df_1.drop(on,axis=1)
	else:
		df_2 = df_2.drop(on,axis=1)

	merged_df = pd.merge(df_1, df_2,how=how, on=temp_list )

	merged_df = merged_df.drop(temp_list, axis=1)

	return merged_df

def fill_header_period(df,fill_period='实销-个月'):
	#转置表之后string类型的表头无法正确排序，用数字型表头转置完成后，再填充周期的文字 比如填充实销1个月，第4周
	before_str = fill_period.split('-')[0]
	after_str = fill_period.split('-')[1]

	new_header = []
	for c in df.columns:
		if type(c) != str or re.match('\d+',c) != None:
			new_header.append( before_str + str(int(c)) + after_str)
		else:
			new_header.append(c)
	df.columns = new_header
	return df

def get_year_month(string):
	"""检查"""
	if type(string) ==str and re.search('\d{4}[年\- /]\d{1,2}[月\- /]',string) != None:
		year_month = re.findall('\d{1,4}',string)
		return [int(x) for x in year_month]     
	return None 

def process_enddate(df):
	#处理截止时间,如果是X年X月的STR格式就提取年和月组合成日期，其他情况采用pd.to_datetime尝试转换
	df['enddate'] = df['enddate']\
				.apply(lambda x: datetime.datetime(get_year_month(x)[0],get_year_month(x)[1],1) \
					if get_year_month(x) != None else pd.to_datetime(x))
	#变成最后一天为截止日期
	df['enddate'] = df['enddate'].apply(lambda x:last_day_of_month(x))
	return df

def split2multi_tables(df,index_list,column_list):
	"""将一个表根据列维度拆分成多个表
	:param df : input df table 
	:param index_list : 
	:param column_list :
	"""
	df_list = [ ]
	for c in columns:
		df_copy = df.loc[:,index_list + column_list]

def pivot2multi_tables(df,index,columns,value_list,filt_column=None,filt_target=None,fillna=None):
	"""将一个需要分N批次保存多个列数值的表摊开成N个
	:param index :same as pd.pivot index, regarding to the fix index column
	:param columns: the original column that needs to be expanded to the header row
	:param  value_list : the rest value columns corresponding to pd param values 
	"""
	#转置，N个渠道摊开，日期放上面一行
	df_copy = df.copy()
	df_list =  []
	#特定字段只提取某个值
	prefix = ''
	if all([filt_column,filt_target]):
		if isinstance(filt_target,list): 
			df_copy = df_copy.loc[df[filt_column].isin(filt_target),:]
		else: 
			df_copy = df_copy.loc[df[filt_column]==filt_target,:]
		prefix = filt_target + '-'
		#注意过滤后过滤的目标字段已经剩下唯一值，故要把过滤字段剔除
		df_copy = df_copy.drop(filt_column,axis=1)

	for values in value_list:
		df_temp = df_copy.pivot(index=index,columns=columns,values=values)

		df_temp.index.name = prefix + values
		if fillna != None :
			df_temp = df_temp.fillna(value=fillna)
		df_list.append(df_temp)

	return df_list

def delete_unnamed_behind(df):
	original_columns = list(df.columns) 
	#去掉unnamed间隔以及间隔后面的所有列
	unnamed_columns =[ x for x in original_columns if \
			 (type(x) == str and re.search('unnamed',x,flags=re.I) != None)] 

	if unnamed_columns :
		end_index = original_columns.index(unnamed_columns[0])
		df = df.iloc[:,:end_index]
	return df 


#删除重复列
def remove_duplicate_columns(df):
	columns = [x for x in df.columns if re.search('\.\d{1,3}$',x) == None]
	df = df.loc[:,columns]
	return df

def check_abnormal_dates(df_worksheet,date_column,table_path,sheet):
	"""
	检查日期是否为空，日期是否填写异常不能正常转换成pandas的datetime
	"""
	#转换发帖时间字段, 可能出现日期错误
	check_date_error = 0 

	check_date_empty = df_worksheet.loc[df_worksheet[date_column].isna()==True,:]
	#记录日期为空的部分，不允许填空
	if not check_date_empty.empty:
		empty_date_row = [str(x+2) for x in check_date_empty.index]
		empty_date_row = ','.join(empty_date_row)
		print('日期不允许为空：文档:{},来源Sheet:{},第{}行，请补充以上日期再运行'.format(table_path,sheet,empty_date_row))
		input('回车键退出')
		sys.exit()

	#如果发帖时间 被转成了int,float格式的时间 （在EXCEL会显示 1月2日，点进去单元格会出现正确的时间格式）
	df_worksheet[date_column] = df_worksheet[date_column].apply(lambda x: \
		datetime.strftime(xldate_as_datetime(x,0),'%Y-%m-%d') if (type(x) == int or type(x)==float) else x)

	try:
		df_worksheet[date_column] = pd.to_datetime(df_worksheet[date_column])
		check_date_error += 1 
	except (ValueError,OutOfBoundsDatetime) as e :      
		for i,t in zip(df_worksheet.index,df_worksheet[date_column]):
			try:
				convert_t = pd.to_datetime(t)
			except ValueError:  
				print('错误日期格式："{}"'.format(t),'文档第{}行'.format(i+2))
	#找不到发帖时间 
	except KeyError:
		print('无法映射出"{}"字段，请检查该字段对应表："{}","{}"'.format(date_column,table_path,sheet))
		input('回车键退出') 

	if check_date_error < 1:
		print('日期格式错误："{}", "{}"，请修改以上日期再运行'.format(table_path,sheet))
		input('回车键退出')
		sys.exit()

	return df_worksheet


def normalize_column_dates(df_worksheet,date_columns):
	"""
	日期是否填写异常不能正常转换成pandas的datetime
	"""
	if type(date_columns) != list:
		date_columns = [date_columns]

	for date_column in date_columns:
		#转换发帖时间字段, 可能出现日期错误

		try : #如果没问题 直接转成datetime 
			df_worksheet[date_column] = pd.to_datetime(df_worksheet[date_column])
		except :
			#如果发帖时间 被转成了int,float格式的时间 （在EXCEL会显示 1月2日，点进去单元格会出现正确的时间格式）
			df_worksheet[date_column] = df_worksheet[date_column].apply(lambda x: \
				datetime.strftime(xldate_as_datetime(x,0),'%Y-%m-%d') if (type(x) == int or type(x)==float) else x)
			try:
				df_worksheet[date_column] = pd.to_datetime(df_worksheet[date_column])
			except :      
				pass

	return df_worksheet

def normalize_dates(df_worksheet,date_columns,table_path='默认',sheet='默认'):
	"""
	检查日期是否为空，日期是否填写异常不能正常转换成pandas的datetime
	"""
	if type(date_columns) != list:
		date_columns = [date_columns]

	for date_column in date_columns:
		#转换发帖时间字段, 可能出现日期错误
		check_date_error = 0 

		check_date_empty = df_worksheet.loc[df_worksheet[date_column].isna()==True,:]
		#记录日期为空的部分，不允许填空
		if not check_date_empty.empty:
			empty_date_row = [str(x+2) for x in check_date_empty.index]
			empty_date_row = ','.join(empty_date_row)
			enter_exit('Date must not be empty：File:{},Sheet:{},Row:{},Please add an valid date then try again!'.format(table_path,sheet,empty_date_row))

		try : #如果没问题 直接转成datetime 
			df_worksheet[date_column] = pd.to_datetime(df_worksheet[date_column])

		except :
			#如果发帖时间 被转成了int,float格式的时间 （在EXCEL会显示 1月2日，点进去单元格会出现正确的时间格式）
			df_worksheet[date_column] = df_worksheet[date_column].apply(lambda x: \
				datetime.datetime.strftime(xldate_as_datetime(x,0),'%Y-%m-%d') if (type(x) == int or type(x)==float) else x)
			try:
				df_worksheet[date_column] = pd.to_datetime(df_worksheet[date_column])
				check_date_error += 1 
			except (ValueError,OutOfBoundsDatetime) as e :      
				for i,t in zip(df_worksheet.index,df_worksheet[date_column]):
					try:
						convert_t = pd.to_datetime(t)
					except ValueError:  
						print('错误日期格式："{}"'.format(t),'文档第{}行'.format(i+2))
			#找不到发帖时间 
			except KeyError:
				print('无法映射出"{}"字段，请检查该字段对应表："{}","{}"'.format(date_column,table_path,sheet))
				input('回车键退出') 

			if check_date_error < 1:
				print('日期格式错误："{}", "{}"，请修改以上日期再运行'.format(table_path,sheet))
				input('回车键退出')
				sys.exit()

	return df_worksheet

def normalize_dates_single(date):
	#上面的简单版，针对单个日期的标准化
	try:
		date = pd.to_datetime(date)
	except:
		if (type(date) == int or type(date)==float):
			date = datetime.datetime.strftime(xldate_as_datetime(x,0),'%Y-%m-%d')
			try:
				pd.to_datetime(date)
			except :
				pd.to_datetime(date)
		else:
			enter_exit(f'Error when converting "{date}"" to datetime')

	return date


def copy_seperate_header_columns(df,pattern):
	"""
	外销国家字段对应.xlsx，把共同的国家部分拆分成多个相同列
	:param df :
	:param pattern :re pattern   '[A-Za-z]{2}'
	"""
	df_copy = df.copy()
	level_0_columns = df_copy.columns.get_level_values(0)

	for c in level_0_columns:
		match_list = re.findall(pattern,c)
		#如果匹配到多个目标，将目标拆分成多个字段，含有相同的数据，但表头变成拆分后的表头结果
		if len(match_list) > 1 :
			for m in match_list:
				#需要将多列分别赋值
				df_copy_need_copy = df_copy.loc[:,df_copy.columns.get_level_values(0)==c]
				for each_col in df_copy_need_copy.columns.tolist():
					df_copy[(m,each_col[1])] = df_copy[(c,each_col[1])]
			#删除Multi index
			df_copy = df_copy.drop([(each_col)],axis=1)
	return df_copy

def normalize_multi_header(df):
	"""将有MultiIndex的column字符串做标准化处理，去掉两边空格等"""
	df_copy = df.copy()
	df_copy_columns = [ tuple(y.strip().lower() for y in x) for x in df_copy.columns ]
	df_copy.columns = pd.core.index.MultiIndex.from_tuples(df_copy_columns)
	return df_copy

def regex_split_column(df_series, split_pat, split_symbol):
	df_series = df_series.apply(lambda x: re.split(split_pat, x))
	def remove_null(lst):
		return [ x for x in lst if x != '' and x != split_symbol ]
	df_series = df_series.apply(remove_null)

	return df_series

def get_flashtext_processor(keyword_list):
	kw_processor = KeywordProcessor()
	for k in keyword_list:
		kw_processor.add_keyword(k)

	return kw_processor

def split_column_by_words(df, split_words_dict,mode='split'):
	mode = mode.strip().lower()
	#split_config_dict {'需要拆分的字段':[ 用来做拆分标准词组的列表 ]}
	for column, split_standard_list in split_words_dict.items():
		print(f'Splitting column:{column},length of standard keywords:{len(split_standard_list[0])},mode:{mode}',)
		#先获取到分隔符
		split_symbol = split_standard_list[1]
		# #全角转半角。去掉重复符号
		split_standard_list = [ normalize_punctuations(x) for x in split_standard_list[0] ] 
		#构造pattern之前 将re的特殊符号转成re可识别的
		split_standard_list = [ replace_re_special(x) for x in split_standard_list ]

		split_standard_list = sorted(split_standard_list, key = lambda x : len(x), reverse=True)

		split_pat = '(' + '|'.join(split_standard_list) +  ')'

		find_lack_columns(df, [column], 'Split column')

		split_pat = re.compile(split_pat,flags=re.I)

		if split_standard_list:
			df['original'] = df[column].copy()
			#符号标准化，全角转半角
			df[column] = df[column].fillna('').astype(str).apply(lambda x: normalize_punctuations(x))

			#区分本来就是空的和非空的,否则会把原本是空的列删除
			df_column_null = df.loc[df[column]=='',:]
			df_column_notnull = df.loc[df[column]!='',:]
			
			if 'split' in mode :
				df_column_notnull[column] = regex_split_column(df_column_notnull[column],split_pat, split_symbol)
			elif 'findall' in mode:
				df_column_notnull[column] = df_column_notnull[column].apply(lambda x: re.findall(split_pat, x))
			elif 'fast' in mode or 'flash' in mode:
				kw_processor = get_flashtext_processor(split_standard_list)
				df_column_notnull[column] = df_column_notnull[column].apply(lambda x: kw_processor.extract_keywords(x) )
			else:  #默认模式用split，如果没有出现太长的关键词或者pattern这个是最完整的
				df_column_notnull[column] = regex_split_column(df_column_notnull[column],split_pat, split_symbol)
				
			df_column_null = stack_list_column(df_column_null, column)
			#重新合并
			df = pd.concat([df_column_null,df_column_notnull],axis=0,ignore_index=True)
			df = stack_list_column(df, column)
		#再用split_symbol拆一次
		elif split_symbol != '':
			df[column] = df[column].apply(lambda x: x.split(split_symbol))
			df = stack_list_column(df, column)

	return df


#将某个表的某个字段通过某个符号拆分之后组成叠成新的表(除了被拆分的字段，其他字段需要复制同样内容新生成一条)
def stack_list_column(df,split_column):
	"""
	:param df : dataframe to be processed 
	:param split_column : list, column that need to be split and stack
	:param split_symbol : symbol(sign) used to split the specific column
	:return stacked Dataframe, specific list type column becomes string on each new row 
	"""
	#防止误修改原始表
	df_copy = df.copy()
	column_list = df_copy.columns.tolist()
	
	#准备构建新的DataFrame
	df_copy_list = []

	symbol_index = column_list.index(split_column)

	for index, value_list in zip(df_copy[split_column].index,df_copy[split_column].values):
		#不含有逗号的部分  和前面含有逗号处理方式 不一样，会存在 有类别的没有现象的情况,所以不能过滤value<1的
		insert_list = df_copy.loc[index,:].tolist()
		if value_list: #如果拆分的该列为空也一样要保留这行数据，不能直接删除！
			for value in value_list:  
				insert_list_copy = insert_list.copy()
				insert_list_copy[symbol_index] = value
				df_copy_list.append(insert_list_copy)
		else:
			insert_list_copy = insert_list.copy()
			insert_list_copy[symbol_index] = ''
			df_copy_list.append(insert_list_copy)
	
	#重新组合DF
	df_copy = pd.DataFrame(df_copy_list,columns= column_list)
	return df_copy

def expand_stacked_column_to_list(df,expand_column,unique_key):
	"""
	#为了统计蓝色部分字段，把故障类别堆积回到列表的状态
	:param df : 原语言拆分匹配后已经被展开的表格
	:param expand_column : 被展开过的字段
	:param unique_key : 唯一编码字段
	"""
	original_columns = df.columns.tolist()
	new_expand_column_name = expand_column + '_列表'
	
	df_copy = df.copy()
	
	df_convert_list = []
	unique_dict = defaultdict(list)
	empty_df = pd.DataFrame()
	
	for index,original_row in df_copy.iterrows():  #读取每一行数据
		_id = original_row[unique_key]
		value = original_row[expand_column]
		
		if unique_dict.get(_id,empty_df).empty:
			original_row[new_expand_column_name] = [value]
			unique_dict[_id] = original_row            
		else:     
			unique_dict[_id][new_expand_column_name].append(value)

	record_list = [x[1].values for x in unique_dict.items()]

	record_df = pd.DataFrame(record_list)
	record_df.columns = original_columns + [new_expand_column_name]
	
	return record_df

def expand_to_columns_pd(df):
	import gc 
	"""
	把相同id的结果 叠到同一行,填入的只能是两列字段，一个是唯一字段，另一个是需要叠成一行并形成不同columns的字段
	默认叠第二个字段
	"""
	original_columns = df.columns
	unique_key = original_columns[0]
	expand_column = original_columns[1]
	
	df_dict = defaultdict(set)
	
	for rows in df.iterrows():
		index = rows[0]
		rows_tuple = rows[1]
		
		_id = rows_tuple[unique_key]
		phen = rows_tuple[expand_column]

		df_dict[_id].add(phen)

	#形成字典后做一个中间格式，然后通过pandas再读取出来
	temp_path = 'temp_dict.txt'

	#计算ECC分类最多的ID有多少列数据
	maximum_ecc_column = 0 
	content = ''

	for k,v in df_dict.items() :
		len_v = len(v)
		if len_v >= maximum_ecc_column:
			maximum_ecc_column = len_v 
		content += k + '\t' + '\t'.join(v) + '\n'
			
	with open(temp_path,'w',encoding='utf-8') as f :
		f.write(content)
		
	del f 
	gc.collect()
	#根据最多的列决定如何读取table 
	new_columns = [unique_key] + [ expand_column + ' ' + str(x) for x in range(1,maximum_ecc_column + 1 ) ]
	#再把数据读回来
	record_df = pd.read_csv(temp_path,header=None,sep='\t', names=new_columns)  
	#删除不需要的临时文件
	os.remove(temp_path) 
	return record_df 


def get_list_partial_sorted(lst,number_regex='([^0-9]+)[0-9]{1,2}'):
	#获取一个列表中按照顺序排列的最后一组的部分，比如 [1,2,1,2,3,4],获取到后面的[1,2,3,4]
	previous = re.match(number_regex,[lst][0]).group()#默认第一个数字是最小值
	store_list = []
	for i in lst[1:]:
		i_num = int(re.match(number_regex,i).group())
		if i_num <= previous:
			store_list = []
			continue
		else:
			store_list.append(i)
	return store_list

def stack_columns_to_multi_row(df,target_stack_name=None,regex_foramt=None):
	"""将多个故障摊开的列以_id为主键，摊开成多行,
	默认最后几列表头中带有0,1,2,3之类的属于应该被叠起来的故障现象
	:param target_stack_name 转换的目标默认字段名
	:param regex_foramt
	return result df 
	"""
	#检查哪几个列带有数字的后缀
	stack_column_pat = '([^0-9]+)[0-9]{1,2}'
	
	df_columns = df.columns
	stack_columns = []

	if regex_foramt == None:
		stack_columns = [x for x in df_columns if re.match(stack_column_pat,x) != None]
	else:
		stack_columns = [x for x in df_columns if re.match(stack_column_pat,x) != None \
		and re.match(regex_foramt) != None]

		if target_stack_name == None and stack_columns:
			target_stack_name = re.match(regex_foramt,stack_columns[0]).group()

	#判断最后一个按顺序排列的列表
	stack_columns = get_list_partial_sorted(stack_columns)

	stack_number_list = [ re.match(stack_column_pat,x).group() for x in df_columns if re.match(stack_column_pat,x) != None \
		and re.match(regex_foramt) != None ]
	#保存数据
	stack_line_list = [ ]
	if len(stack_columns) < 2:
		enter_exit('电商爬虫数据中没有找到任何带有数字后缀的多个故障现象列')
	else:  #如果找到了对应列
		stack_header_name = re.match(stack_column_pat,stack_columns[0]).group(1)
		for i,row in df.loc[:,stack_columns].iterrows():
			stack_line = ','.join(sorted([x for x in row.values if type(x)==str and x !='nan']))
			stack_line_list.append(stack_line)

	df[target_stack_name] = stack_line_list
	df_columns = [x for x in df_columns if x not in stack_columns] + [target_stack_name]

	#堆叠字段全部放后面
	df = df.loc[:,df_columns]
	return  df 


def get_target_sheet(path,target_name):
	"""
	通过Sheet名称获取目标Sheet，不读取隐藏的Sheet, 出现同名不同日期的Sheet优先读取后面最新的
	"""
	match_list = [ ]
	df_workbook = pd.ExcelFile(path)
	sheets_property_list = df_workbook.book.sheets()

	for sheet_property in sheets_property_list:
		sheet_name = sheet_property.name 
		sheet_visibility = sheet_property.visibility 
		if sheet_visibility ==  0  and target_name in sheet_name:
			match_list.append(sheet_name)
	try:
		match_sheetname = match_list[-1]
	except IndexError:
		print('找不到名称为"{}"的Sheet'.format(target_name))
		return None

	if len(match_list) > 1 :
		 print('\"{0}\"存在{1}份sheet文档,正在读取{2}'.format(target_name,len(match_list),match_sheetname))
	#遇到重复字段 overwrite 覆盖掉 -----mangle_dupe_cols not supported yet...还没支持
	# df = df_workbook.parse(match_sheetname,mangle_dupe_cols=False)
	df = df_workbook.parse(match_sheetname)
	df.columns = [x.strip() for x in df.columns]
	#关闭
	# df_workbook.close()
	df_workbook.close()
	return df

def get_target_sheet_wb(workbook,target_name,header=None):
	"""
	通过Sheet名称获取目标Sheet，不读取隐藏的Sheet, 出现同名不同日期的Sheet优先读取后面最新的
	和get_target_sheet的区别在于传入的是一个workbook，并且最后不会关闭workbook（需要读取其他sheet）
	"""
	match_list = [ ]
	sheets_property_list = workbook.book.sheets()

	for sheet_property in sheets_property_list:
		sheet_name = sheet_property.name 
		sheet_visibility = sheet_property.visibility 
		if sheet_visibility ==  0  and target_name in sheet_name:
			match_list.append(sheet_name)
	try:
		match_sheetname = match_list[-1]
	except IndexError:
		print('找不到名称为"{}"的Sheet'.format(target_name))
		return None

	if len(match_list) > 1 :
		 print('\"{0}\"存在{1}份sheet文档,正在读取{2}'.format(target_name,len(match_list),match_sheetname))
	#遇到重复字段 overwrite 覆盖掉 -----mangle_dupe_cols not supported yet...还没支持
	# df = df_workbook.parse(match_sheetname,mangle_dupe_cols=False)
	if header != None:
		df = df_workbook.parse(match_sheetname,header=header)
	else:
		df = df_workbook.parse(match_sheetname)

	df.columns = [x.strip() for x in df.columns]
	return df

def convert_word_freq_dict(dict_text):
	text = ''
	if type(dict_text) == str:
		dict_text = json.loads(dict_text)
		for k,v in json.items():
			for i in range(v):
				text += ' ' + k

		return text.strip()
	else:
		return dict_text

def group_by_concat(df, group_column, agg_func_column):

	#防止改到另外的部分
	df_copy = df.copy()
	df_copy[agg_func_column] = df_copy[agg_func_column].fillna('').apply(lambda x : x if type(x) == str and x != '' else str(x) )
	df_agg = df_copy.groupby(group_column)[agg_func_column].apply(' '.join)
	#以上结果是一个Series,需要转成DF, 不在这里reset_index，后面append本次结果 后面会用到groupby的index来concat
	df_agg = df_agg.to_frame()

	return df_agg 

def group_word_count(df, group_column, agg_func, agg_func_column, keyword_list, stopword_list, count_keywords_only=False):
	#获取分组词频, 每行单词不去重
	df_agg = group_by_concat(df,group_column,agg_func_column)
	#进行分词和统计等处理, 使用swifter加速
	df_agg[agg_func_column] = df_agg[agg_func_column].swifter.progress_bar(enable=True, desc=agg_func)\
						.apply(lambda x: get_word_freq_dict(x, keyword_list,stopword_list,
													count_keywords_only=count_keywords_only, word_num=200))

	df_agg[agg_func_column] = df_agg[agg_func_column].apply(lambda x: DfDict(x) )
	#转成json,文本格式
	# df_agg[agg_func_column] = df_agg[agg_func_column].apply(lambda x: json.dumps(x))
	return df_agg

def group_dict_sum(df, group_column, agg_func_column):

	df[agg_func_column] = df[agg_func_column].apply(lambda x: json.loads(x) if type(x) == str else x )

	df_agg = df.groupby(group_column).agg('sum')

	#转成json,文本格式
	df_agg[agg_func_column] = df_agg[agg_func_column].apply(lambda x: json.dumps(x,ensure_ascii=False) if isinstance(x,dict) else r'{}' )

	return df_agg

def word_agg_func(df, group_column, agg_func, agg_func_column, keyword_list, stopword_list):

	#如果是统计的关键词，停用词直接置空
	if 'keyword' in agg_func :
		stopword_list = [ ]
		#不允许关键词为空
		if len(keyword_list) == 0 :
			enter_exit('Can not find keywords with empty keyword list,please check the keyword files !') 

	#每条出现多次，只统计一次,不适用与sum的逻辑
	if 'unique' in agg_func and 'count' in agg_func:
		df[agg_func_column] = df[agg_func_column].swifter.progress_bar(enable=True, desc=agg_func).apply(
								lambda x: ' '.join(set(process_text_eng(x, keyword_list, stopword_list))) if type(x) == str else x)

	if 'keyword_count_en' == agg_func[:len('keyword_count_en')]:
		df_agg = group_word_count(df, group_column, agg_func, agg_func_column, keyword_list=keyword_list, 
									stopword_list = [ ], count_keywords_only=True)

	elif  "word_count_en" == agg_func[:len('word_count_en')]:
		df_agg = group_word_count(df, group_column, agg_func, agg_func_column, keyword_list, stopword_list,
									count_keywords_only=False)

	elif 'keyword_sum' == agg_func[:len('keyword_sum')]:  #和word_sum一样计算 字典内的数字相加
		df_agg = group_dict_sum(df, group_column, agg_func_column )

	elif 'word_sum' == agg_func[:len('word_sum')]:
		df_agg = group_dict_sum(df, group_column, agg_func_column )

	else: #以下是暂时复制的，中文分词还没写
		enter_exit(f'{agg_func} not supported yet.')

	return df_agg 


def create_group_empty(df, groups, value_column, fillna=0):
	agg_df_sub = df.loc[:,groups + [ value_column ] ] 
	agg_df_sub = agg_df_sub.drop_duplicates(subset=groups)
	agg_df_sub = agg_df_sub.set_index(groups)
	agg_df_sub[value_column] = fillna

	return agg_df_sub 

def group_basic_agg(df,group_column,agg_func, value_column=None, keyword_list= [],stopword_list=[ ], group_index=False):
	"""
	根据分类统计基本的类别数量
	:param df : input df 
	:param group_column : groupby column, list
	:param agg_function : same as df agg func
	:return :  dataframe contains agg result
	"""
	df_copy = df.copy()

	if type(group_column) != list:
		group_column = [ group_column ]

	#是否有传入指定数值的列,不传入只能做简单的count统计
	if value_column == None:
		agg_func_column = [ agg_func ]
		df_copy[agg_func_column] = True
	else:
		agg_func_column =  [ value_column ] 

	lack_column_list = find_lack_columns(df_copy, agg_func_column + group_column, 'Group statistic')

	df_copy = df_copy.loc[:, agg_func_column + group_column]

	df_copy = df_copy.fillna(0)

	if 'word' in agg_func:
		#注意这里的agg_func_columns是一个列表，如果需要做词频统计，需要取出第一个
		df_copy_agg = word_agg_func(df_copy, group_column, agg_func, agg_func_column[0],keyword_list,stopword_list)
	else:
		try:
			df_copy_agg =  df_copy.groupby(group_column).agg(agg_func).sort_values(by=agg_func_column,ascending=False).fillna(0)
		except AttributeError:
			enter_exit(f'"{agg_func}" function not found')
		except: 
			enter_exit(f'Failed to execute calc function: "{agg_func}" in "{agg_func_column[0]}"')

	if not group_index :
		df_copy_agg = df_copy_agg.reset_index()
	
	return df_copy_agg


def calc_total(df,total_column=0):
	"""最后计算一行总计数量
	:param df : input df 
	:return : new df that has the last row name total 
	"""
	df_copy = df.copy()
	df_copy = df_copy.fillna(value=0).reset_index()
	#calculate total get the last index
	# last_index = df_copy.index.values[-1] + 1 
	last_index = df_copy.shape[0]
	total_series = pd.Series(df_copy.sum(axis=0,numeric_only=True),name=last_index)
	# total_series 
	df_copy = df_copy.append(total_series)
	if type(total_column) == int :
		df_columns = df.columns.tolist()
		total_column = df_columns[total_column]

	df_copy[total_column][last_index] = 'total'
	
	return df_copy


def calc_percent(df,total_column,calc_columns=None):
	"""
	每个数字列的都计算一个比率, 默认左边第一列是文字类型的列，默认不计算非数字类型的列
	:param df : 含有各个类别统计数值的表格，基本格式：第一列是类型，第二列是每个类型的数量，
			   后面的列都是不同条件的数量，最后一行是总计
	:param total_column : 指定的总计列，会通过总计列决定从哪里开始计算每一列的比率
	:param calc_columns : 指定哪些列需要计算百分比结果
	:return new df with percentage counted new columns, if df is empty return None 
	"""
	if not df.empty:
		df_copy = df.copy()
		original_columns = list(df_copy.columns)
		#默认计算所有的数字类型列(numerical)的百分比
		if  calc_columns == None :
			calc_columns = original_columns

		elif type(calc_columns) != list :
			calc_columns = [calc_columns]

		#去掉非数字类型的列
		numerical_cols =  df_copy.select_dtypes(exclude=['object','datetime']).columns

		calc_columns = [x for x in calc_columns if x in numerical_cols]

		for c in calc_columns:
			#如果是总数比例，分母不同，而且需要去掉最后一行的总计再SUM 
			if c == total_column:
				denominator = df_copy[c].iloc[:-1].sum()
			else:
				denominator = df_copy[total_column]

			columns = list(df_copy.columns)
			c_index = columns.index(c)
			series_name = c + '-pct'
			percent_series = df_copy[c]*100 / denominator

			#注意pandas的insert不返回任何结果
			df_copy.insert(c_index+1, series_name, percent_series)
			#格式化展示比率
			# df_copy[series_name] = df_copy[series_name].apply(lambda x: '{}%'.format(round(x,2)) if type(x)==float else x )
		return df_copy
	else:
		return None 

def get_2_columns_div(df,dividend,divisor):
	"""
	计算一个表格两列字段相处的结果，并且将结果round两位小数点填充百分百展示
	:param df : input df 
	:param dividend : 被除数  分子
	:param divisor :除数   分母
	"""
	result_series = (df[dividend]*100/df[divisor]).apply(lambda x: '{0:0.2f}%'.format(x) if type(x)==float else x )

	return result_series


def get_outer_join_diff(df_list,on=None,left_right=None):
	"""
	参考：http://www.datasciencemadesimple.com/join-merge-data-frames-pandas-python/
	Pandas的outer join函数默认是full outer join,即返回的是全部两边都有的数据，求不相交的部分需要进行以下操作
	获取outer join结果集 减去 inner join 结果集 取得不相同的数据
	
	:param df list :  list containing two dataframes 
	:param on : columns that used to join on 
	:param left_right : indicates the name of each df in the df_list,will be added to the new_column nams 
	return 两个表不相交的集合
	
	"""
	#防止修改到原表格
	df_1_copy = df_list[0].copy()
	df_2_copy = df_list[1].copy()
	#如果关联字段没有填，默认关联两个表之间的相同
	if on == None:
		columns_1  = set(df_1_copy.columns)
		columns_2  = set(df_1_copy.columns)
		on = list(columns_1 & columns_2)

	if type(df_list) != list or len(df_list) != 2:
		raise ValueError("df_list must be of list type and contain 2 dataframes")
	#outer join是包含有所有结果集的，只是匹配不上关联字段的 第三个字段会出现空，需要确保第三个字段原本是非空才好进行判定
	if left_right == None:
		left_right = ['Left','Right']
		
	df_1_copy['table_source_1'] = left_right[0]
	df_2_copy['table_source_2'] = left_right[1]
	
	outer_join_df = pd.merge(df_1_copy,df_2_copy,'outer',on=on)
	#此时两个表里面的join_check_x和join_check_y一旦出现空就是要求的非交集
	outer_join_diff_df = outer_join_df.loc[(outer_join_df['table_source_1'].isna()==True)|(outer_join_df['table_source_2'].isna()==True),:]
	#填充构建一个新的合并来源列
	outer_join_diff_df['table_source_1'].fillna(value=outer_join_diff_df['table_source_2'],inplace=True)
	outer_join_diff_df = outer_join_diff_df.rename({'table_source_1':'from'},axis=1)
	#删除已经填充过去的列
	del outer_join_diff_df['table_source_2']
	
	return outer_join_diff_df

def fillna_by_column(df,same_column_name):
	"""
	填充经过关联后 带有_x和_y的同名字段, 用右边填充左边的数据
	:param df 
	:param same_column_name 
	"""
	df_copy = df.copy()
	column_x = same_column_name+'_x'
	column_y = same_column_name+'_y'
	
	df_copy[column_x] = df_copy[column_x].fillna(df_copy[column_y])
	del df_copy[column_y]
	df_copy = df_copy.rename({column_x:same_column_name},axis=1)
	
	return df_copy


def get_rank_column(df,  rank_group):
	#默认需要排序ID 
	df = df.sort_values(by=rank_group)

	df['rank'] = df.groupby(rank_group).cumcount() + 1 

	return df 


import traceback 
import logging 
import datetime 
from .os_functions import enter_exit
from .excel_functions import write_format_columns
from pandas import DataFrame

#装饰器 可以放多个在一个函数上面

def catch_and_print(func):
	def wrapper(*args, **kwargs):
		try:
			return func(*args, **kwargs)
		except Exception as e :
			logging.error(traceback.format_exc())
			enter_exit(f'Error: Calling function: {func.__name__}')

	return wrapper

def df_row_num_decorator(func):
    def wrapper(*args, **kwargs):
        func_result = func(*args, **kwargs)
        if type(func_result) != tuple:
        	df_num = func_result.shape[0]
        else:
        	df_num = func_result[0].shape[0]

        print(f'Amount of data rows:{df_num}')
        #必须要返回func_result的原始结果，否则原始函数的结果会变成wrapper的结果
        return func_result
    return wrapper

def get_run_time(func):
    t_start = datetime.datetime.now()
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        t_stop = datetime.datetime.now()
        print('Time used:', round((t_stop - t_start).total_seconds(),1),'seconds')
        return result
    return wrapper





import re 
import nbformat
from nbconvert import PythonExporter
import warnings 

warnings.filterwarnings('ignore')

def nbconverter(notebookPath,directionPath=None):
	"""method name is the same as jupyter default converter name :
	   Nbconvert(notebookPath,directionPath) 
	:param notebookPath: source path of the ipynb file you want to convert
	:param direction Path : direction path of the formatted .py file 
	"""
	try:
		with open(notebookPath,'r',encoding='utf-8') as fh:
			nb = nbformat.reads(fh.read(), nbformat.NO_CONVERT)
	except FileNotFoundError:
		if '.ipynb' not in notebookPath:
			notebookPath +=  '.ipynb'
		with open(notebookPath,'r',encoding='utf-8') as fh:
			nb = nbformat.reads(fh.read(), nbformat.NO_CONVERT)

	if directionPath == None:
		directionPath = notebookPath.replace('.ipynb','.py')

	pattern_input = r'# In\[[\d\s]*\]:'
	pattern_comment = '^#.+'
	pattern_variables = r'^[a-zA-Z0-9_]+\[?[ |0-9]*\]? *$'
	pattern_square_bracket = r'^\[.*\]$'
	pattern_string = r"^[\'|\"].*[\'|\"] *$"
	pattern_number = r'^\d+ *[\+|\-|\*|\/]? *\d* *$'

	def print_match(matched):
		return 'print({})'.format(matched.group().strip())

	exporter =     PythonExporter()
	source, meta = exporter.from_notebook_node(nb)

	source = source.split('\n')[2:]
	source = [ t for t in source if t and not re.match(pattern_input,t)]
	source = [ '\n' + t if re.match(pattern_comment,t) else t for t in source ]
	source = [re.sub(pattern=pattern_variables,repl=print_match,string=t) for t in source]
	source = [re.sub(pattern=pattern_square_bracket,repl=print_match,string=t) for t in source]
	source = [re.sub(pattern=pattern_string,repl=print_match,string=t) for t in source]

	source = '\n'.join(source) + '\n\n'
	with open(directionPath, 'w+',encoding='utf-8') as fh:
		fh.write(source)
		print('{} has been saved'.format(directionPath))



nbconverter(r"C:\Users\Administrator.DG-11030335\Scripts\voc_alarm\外销意见反馈预警监控\convert2alarm.ipynb")